# Copyright 2016 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from elasticsearch import exceptions as es_exceptions
from oslo_config import cfg
import six
from unittest import mock

from searchlight.elasticsearch.plugins import helper
from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight.tests import fake_plugins
from searchlight.tests import utils as test_utils


class TestIndexingHelper(test_utils.BaseTestCase):
    def test_role_separated_save_docs(self):
        """Test admin only fields are correctly removed from serialization
        and that alternate _id values are used
        """
        mock_engine = mock.Mock()
        plugin = fake_plugins.RoleSeparatedPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            count = len(plugin.get_objects())
            fake_versions = range(1, count + 1)
            indexing_helper.save_documents(plugin.get_objects(), fake_versions)
            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        self.assertEqual(4, len(actions))
        self.assertEqual(set(['role-fake1_ADMIN', 'role-fake2_ADMIN',
                             'role-fake1_USER', 'role-fake2_USER']),
                         set(action['_id'] for action in actions))
        self.assertEqual(set(fake_versions),
                         set(action['_version'] for action in actions))
        self.assertEqual(['external'] * 4,
                         [action['_version_type'] for action in actions])
        # This plugin filters on admin_wildcard_* and admin_specific
        fake1_admin = list(filter(lambda a: a['_id'] == 'role-fake1_ADMIN',
                                  actions))[0]['_source']
        self.assertEqual('role-fake1', fake1_admin['id'])
        self.assertIn('public_field', fake1_admin)
        self.assertIn('admin_wildcard_this', fake1_admin)
        self.assertEqual('admin', fake1_admin[ROLE_USER_FIELD])

        fake1_user = list(filter(lambda a: a['_id'] == 'role-fake1_USER',
                                 actions))[0]['_source']
        self.assertEqual('role-fake1', fake1_user['id'])
        self.assertIn('public_field', fake1_user)
        self.assertNotIn('admin_wildcard_this', fake1_user)
        self.assertNotIn('admin_specific', fake1_user)
        self.assertEqual('user', fake1_user[ROLE_USER_FIELD])

    def test_non_role_separated_save_docs(self):
        """Test that for a plugin that doesn't specify any protected fields,
        ids are left alone and there's only one copy of each indexed doc
        """
        mock_engine = mock.Mock()
        plugin = fake_plugins.NonRoleSeparatedPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            count = len(plugin.get_objects())
            fake_versions = range(1, count + 1)
            indexing_helper.save_documents(plugin.get_objects(), fake_versions)
            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        self.assertEqual(2, len(actions))
        self.assertEqual(set(['non-role-fake1', 'non-role-fake2']),
                         set(action['_id'] for action in actions))
        self.assertEqual(set(fake_versions),
                         set(action['_version'] for action in actions))
        self.assertEqual(['external'] * 2,
                         [action['_version_type'] for action in actions])
        fake1 = actions[0]['_source']
        self.assertEqual(['admin', 'user'],
                         sorted(fake1.pop(ROLE_USER_FIELD)))
        self.assertEqual(fake_plugins.NON_ROLE_SEPARATED_DATA[0],
                         actions[0]['_source'])

    @mock.patch('searchlight.elasticsearch.plugins.helper.helpers.bulk')
    @mock.patch.object(cfg.CONF, 'resource_plugin')
    @mock.patch.object(cfg.CONF, 'service_credentials')
    def test_region_mapping(self, service_credentials_conf,
                            resource_plugin_conf, mock_bulk):
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSimplePlugin(es_engine=mock_engine)

        resource_plugin_conf.include_region_name = True
        service_credentials_conf.os_region_name = 'test-region'

        indexing_helper = helper.IndexingHelper(plugin)

        _, mapping = six.next(plugin.get_full_mapping())
        self.assertIn('region_name', mapping['properties'])

        count = len(plugin.get_objects())
        fake_versions = range(1, count + 1)
        indexing_helper.save_documents(plugin.get_objects(),
                                       fake_versions)

        self.assertEqual(1, len(mock_bulk.call_args_list))
        actions = list(mock_bulk.call_args_list[0][1]['actions'])
        self.assertEqual(['test-region'],
                         actions[0]['_source']['region_name'])

        # Test without a region
        resource_plugin_conf.include_region_name = False
        mock_bulk.reset_mock()

        _, mapping = six.next(plugin.get_full_mapping())
        self.assertNotIn('region_name', mapping['properties'])
        indexing_helper.save_documents(plugin.get_objects(),
                                       fake_versions)
        actions = list(mock_bulk.call_args_list[0][1]['actions'])
        self.assertNotIn('region_name', actions[0]['_source'])

    def test_role_separated_delete(self):
        """Test that deletion for a role-separated plugin deletes both docs"""
        mock_engine = mock.Mock()
        plugin = fake_plugins.RoleSeparatedPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.delete_document({'_id': 'role-fake1'})

            expected_delete_actions = [
                {'_op_type': 'delete',
                 '_id': 'role-fake1_ADMIN'},
                {'_op_type': 'delete',
                 '_id': 'role-fake1_USER'}
            ]
            mock_bulk.assert_called_once_with(
                client=plugin.engine,
                index=plugin.alias_name_listener,
                doc_type=plugin.document_type,
                actions=expected_delete_actions)

    def test_non_role_separated_delete(self):
        """Test that deletion for a role-separated plugin deletes the doc"""
        mock_engine = mock.Mock()
        plugin = fake_plugins.NonRoleSeparatedPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.delete_document({'_id': 'non-role-fake1'})

            expected_delete_actions = [
                {'_op_type': 'delete',
                 '_id': 'non-role-fake1'}
            ]
            mock_bulk.assert_called_once_with(
                client=plugin.engine,
                index=plugin.alias_name_listener,
                doc_type=plugin.document_type,
                actions=expected_delete_actions)

    def test_save_child_parent_both_separated(self):
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSeparatedChildPlugin(es_engine=mock_engine)
        parent_plugin = fake_plugins.RoleSeparatedPlugin(es_engine=mock_engine)
        plugin.register_parent(parent_plugin)

        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'

        child_docs = copy.deepcopy(fake_plugins.CHILD_DATA)

        # First run where both child and parent are role-separated (and thus
        # we'd expect two copies of both parent and child with appropriate
        # ids linking them)
        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.save_documents(child_docs)

            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        expected_admin_doc = copy.deepcopy(child_docs[0])
        expected_admin_doc[ROLE_USER_FIELD] = 'admin'
        expected_user_doc = copy.deepcopy(child_docs[0])
        expected_user_doc[ROLE_USER_FIELD] = 'user'

        expected_actions = [
            {'_op_type': 'index', '_id': 'child1_ADMIN',
             '_source': expected_admin_doc, '_parent': 'simple1_ADMIN'},
            {'_op_type': 'index', '_id': 'child1_USER',
             '_source': expected_user_doc, '_parent': 'simple1_USER'}
        ]

        self.assertEqual(expected_actions, list(actions))

    @mock.patch.object(fake_plugins.FakeSeparatedChildPlugin,
                       'requires_role_separation',
                       new_callable=mock.PropertyMock)
    def test_save_parent_only_separated(self, mock_role_separated):
        """Test where the parent document is role separated but this child
        is not; it's expected the parent's _USER documents be used for _parent
        """
        mock_role_separated.return_value = False

        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSeparatedChildPlugin(es_engine=mock_engine)
        parent_plugin = fake_plugins.RoleSeparatedPlugin(es_engine=mock_engine)
        plugin.register_parent(parent_plugin)

        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'

        child_docs = copy.deepcopy(fake_plugins.CHILD_DATA)

        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.save_documents(child_docs)

            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        expected_doc = copy.deepcopy(child_docs[0])
        expected_doc[ROLE_USER_FIELD] = ['user', 'admin']

        expected_actions = [
            {'_op_type': 'index', '_id': 'child1',
             '_source': expected_doc, '_parent': 'simple1_USER'},
        ]

        self.assertEqual(expected_actions, list(actions))

    @mock.patch.object(fake_plugins.FakeChildPlugin,
                       'requires_role_separation',
                       new_callable=mock.PropertyMock)
    def test_child_only_separated(self, mock_role_separated):
        """Test where the child (and not the parent) is separated. Expect
        two documents with the same parent
        """
        mock_role_separated.return_value = True

        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeChildPlugin(es_engine=mock_engine)
        parent_plugin = fake_plugins.FakeSimplePlugin(es_engine=mock_engine)
        plugin.register_parent(parent_plugin)

        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'

        child_docs = copy.deepcopy(fake_plugins.CHILD_DATA)

        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.save_documents(child_docs)

            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        expected_admin_doc = copy.deepcopy(child_docs[0])
        expected_admin_doc[ROLE_USER_FIELD] = 'admin'
        expected_user_doc = copy.deepcopy(child_docs[0])
        expected_user_doc[ROLE_USER_FIELD] = 'user'

        expected_actions = [
            {'_op_type': 'index', '_id': 'child1_ADMIN',
             '_source': expected_admin_doc, '_parent': 'simple1'},
            {'_op_type': 'index', '_id': 'child1_USER',
             '_source': expected_user_doc, '_parent': 'simple1'}
        ]
        self.assertEqual(expected_actions, list(actions))

    def test_delete_children_role_separated(self):
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSeparatedChildPlugin(es_engine=mock_engine)
        parent_plugin = fake_plugins.RoleSeparatedPlugin(es_engine=mock_engine)
        plugin.register_parent(parent_plugin)

        indexing_helper = helper.IndexingHelper(plugin)

        scan_name = 'searchlight.elasticsearch.plugins.helper.helpers.scan'
        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'

        mock_scan_data = [
            {'_id': '1_ADMIN',
             'fields': {'_parent': 'p1_ADMIN'}},
            {'_id': '1_USER',
             'fields': {'_parent': 'p1_USER'}}
        ]
        with mock.patch(scan_name, return_value=mock_scan_data) as mock_scan:
            with mock.patch(bulk_name) as mock_bulk:
                indexing_helper.delete_documents_with_parent('p1')

                parent_type = plugin.parent_plugin_type()
                base_full_parent_type = '%s#%s' % (parent_type, 'p1')
                expected_scan_query = {
                    'fields': ['_parent', '_routing'],
                    'query': {
                        'terms': {
                            '_parent': [base_full_parent_type + '_ADMIN',
                                        base_full_parent_type + '_USER']
                        }
                    }
                }
                mock_scan.assert_called_with(
                    client=plugin.engine,
                    index=plugin.alias_name_listener,
                    doc_type=plugin.document_type,
                    query=expected_scan_query
                )

                expected_delete_actions = [
                    {'_op_type': 'delete',
                     '_id': '1_ADMIN',
                     '_parent': 'p1_ADMIN'},
                    {'_op_type': 'delete',
                     '_id': '1_USER',
                     '_parent': 'p1_USER'}
                ]
                mock_bulk.assert_called_with(
                    client=plugin.engine,
                    index=plugin.alias_name_listener,
                    doc_type=plugin.document_type,
                    actions=expected_delete_actions
                )

    def test_delete_children_non_role_separated(self):
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeChildPlugin(es_engine=mock_engine)
        parent_plugin = fake_plugins.FakeSimplePlugin(es_engine=mock_engine)
        plugin.register_parent(parent_plugin)
        indexing_helper = helper.IndexingHelper(plugin)

        scan_name = 'searchlight.elasticsearch.plugins.helper.helpers.scan'
        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'

        mock_scan_data = [
            {'_id': '1',
             'fields': {'_parent': 'p1'}},
        ]
        with mock.patch(scan_name, return_value=mock_scan_data) as mock_scan:
            with mock.patch(bulk_name) as mock_bulk:
                indexing_helper.delete_documents_with_parent('p1')

                parent_type = plugin.parent_plugin_type()
                expected_scan_query = {
                    'fields': ['_parent', '_routing'],
                    'query': {
                        'term': {
                            '_parent': '%s#%s' % (parent_type, 'p1')
                        }
                    }
                }
                mock_scan.assert_called_with(
                    client=plugin.engine,
                    index=plugin.alias_name_listener,
                    doc_type=plugin.document_type,
                    query=expected_scan_query
                )

                expected_delete_actions = [
                    {'_op_type': 'delete',
                     '_id': '1',
                     '_parent': 'p1'}
                ]
                mock_bulk.assert_called_with(
                    client=plugin.engine,
                    index=plugin.alias_name_listener,
                    doc_type=plugin.document_type,
                    actions=expected_delete_actions
                )

    def test_routing_save_docs(self):
        """Test that for a plugin that specifies routing_id field will
        end up with "_routing" set while indexing.
        """
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSimpleRoutingPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            count = len(plugin.get_objects())
            fake_versions = range(1, count + 1)
            indexing_helper.save_documents(plugin.get_objects(), fake_versions)
            self.assertEqual(1, len(mock_bulk.call_args_list))
            actions = list(mock_bulk.call_args_list[0][1]['actions'])

        # '_routing' is added to action only if set in the plugin property
        # FakeSimpleRoutingPlugin has it defined.
        self.assertIs(True, '_routing' in actions[0])
        self.assertEqual('tenant1', actions[0]['_routing'])

    def test_routing_delete(self):
        """Test that deletion for a routing based plugin deletes docs"""
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSimpleRoutingPlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            indexing_helper.delete_document(
                {'_id': 'id_for_routing_plugin-fake1',
                 '_routing': 'tenant1'})

            expected_delete_actions = [
                {'_op_type': 'delete',
                 '_id': 'id_for_routing_plugin-fake1',
                 '_routing': 'tenant1'}
            ]
            mock_bulk.assert_called_once_with(
                client=plugin.engine,
                index=plugin.alias_name_listener,
                doc_type=plugin.document_type,
                actions=expected_delete_actions)

    def test_bulk_index_error_handling(self):
        """Check that 404 and 409 errors are appropriately ignored"""
        from elasticsearch import helpers

        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSimplePlugin(es_engine=mock_engine)
        indexing_helper = helper.IndexingHelper(plugin)

        bulk_name = 'searchlight.elasticsearch.plugins.helper.helpers.bulk'
        with mock.patch(bulk_name) as mock_bulk:
            mock_bulk.side_effect = helpers.BulkIndexError(
                "1 document(s) failed to index",
                [{'delete': {"_id": "1", "error": "Some error", "status": 404,
                             "exception": helpers.TransportError()}}]
            )

            indexing_helper.delete_documents([{'_id': '1'}])

            self.assertEqual(1, mock_bulk.call_count)

        with mock.patch(bulk_name) as mock_bulk:
            mock_bulk.side_effect = helpers.BulkIndexError(
                "1 document(s) failed to index",
                [{'index': {"_id": "1",
                            "error": {
                                "type": "version_conflict_engine_exception"},
                            "status": 409}}]
            )
            indexing_helper.save_documents([{'id': '1'}])
            self.assertEqual(1, mock_bulk.call_count)

    def test_multiple_alias_exception_detection(self):
        alias_exc = es_exceptions.RequestError(
            400, "Something " + helper.ALIAS_EXCEPTION_STRING, {})
        self.assertTrue(helper._is_multiple_alias_exception(alias_exc))

        other_exc = Exception("Blah blah")
        self.assertFalse(helper._is_multiple_alias_exception(other_exc))

    def test_multiple_alias_exception_elasticsearch2(self):
        # The es-2 format is different
        alias_exc = es_exceptions.RequestError(
            400,
            'illegal_argument_exception',
            {"error": {
                "root_cause": [{
                    "type": "illegal_argument_exception",
                    "reason": "Something " + helper.ALIAS_EXCEPTION_STRING
                }],
                "type": "illegal_argument_exception",
                "reason": "Something " + helper.ALIAS_EXCEPTION_STRING
            }})
        self.assertTrue(helper._is_multiple_alias_exception(alias_exc))

    def test_strip_suffix(self):
        _id = 'aaaa-bbbb'
        admin_id = _id + helper.ADMIN_ID_SUFFIX
        user_id = _id + helper.USER_ID_SUFFIX

        # Test does nothing if there's nothing to do
        self.assertEqual(
            _id,
            helper.strip_role_suffix(_id))
        self.assertEqual(
            _id,
            helper.strip_role_suffix(_id, helper.ADMIN_ID_SUFFIX))

        # Test when there is something to do
        self.assertEqual(
            _id,
            helper.strip_role_suffix(admin_id, helper.ADMIN_ID_SUFFIX))
        self.assertEqual(
            _id,
            helper.strip_role_suffix(user_id, helper.USER_ID_SUFFIX))
        self.assertEqual(_id, helper.strip_role_suffix(user_id))
        self.assertEqual(_id, helper.strip_role_suffix(user_id))

        # Test mismatches
        self.assertEqual(
            admin_id,
            helper.strip_role_suffix(admin_id, helper.USER_ID_SUFFIX))
        self.assertEqual(
            user_id,
            helper.strip_role_suffix(user_id, helper.ADMIN_ID_SUFFIX))

    def test_delete_unknown_parent(self):
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeSeparatedChildPlugin(es_engine=mock_engine)
        doc_id = 'aaaa-bbbb'
        parent_id = '1111-2222'

        # Mock the delete function because the internals are tested elsewhere
        # First test with parent and routing
        with mock.patch.object(plugin.index_helper,
                               'delete_document') as mock_del:
            mock_engine.search.return_value = {
                'hits': {
                    'total': 1,
                    'hits': [{'_id': doc_id + helper.ADMIN_ID_SUFFIX,
                              '_routing': parent_id,
                              '_parent': parent_id + helper.ADMIN_ID_SUFFIX}]
                }
            }
            plugin.index_helper.delete_document_unknown_parent(doc_id)
            mock_del.assert_called_with({'_id': doc_id,
                                         '_parent': parent_id,
                                         '_routing': parent_id})

        # Now no explicit routing
        with mock.patch.object(plugin.index_helper,
                               'delete_document') as mock_del:
            mock_engine.search.return_value = {
                'hits': {
                    'total': 1,
                    'hits': [{'_id': doc_id + helper.ADMIN_ID_SUFFIX,
                              '_parent': parent_id + helper.ADMIN_ID_SUFFIX}]
                }
            }
            plugin.index_helper.delete_document_unknown_parent(doc_id)
            mock_del.assert_called_with({'_id': doc_id,
                                         '_parent': parent_id})

        # Test no results found
        with mock.patch.object(plugin.index_helper,
                               'delete_document') as mock_del:
            mock_engine.search.return_value = {
                'hits': {
                    'total': 0,
                    'hits': []
                }
            }
            plugin.index_helper.delete_document_unknown_parent(doc_id)
            self.assertEqual(0, mock_del.call_count)

        # Also test a non-separated plugin
        mock_engine = mock.Mock()
        plugin = fake_plugins.FakeChildPlugin(es_engine=mock_engine)
        with mock.patch.object(plugin.index_helper,
                               'delete_document') as mock_del:
            mock_engine.search.return_value = {
                'hits': {
                    'total': 1,
                    'hits': [{'_id': doc_id,
                              '_parent': parent_id}]
                }
            }
            plugin.index_helper.delete_document_unknown_parent(doc_id)
            mock_del.assert_called_with({'_id': doc_id,
                                         '_parent': parent_id})



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging

from searchlight.common import utils
from searchlight.elasticsearch.plugins import openstack_clients

LOG = logging.getLogger(__name__)

AUTH_PREFIX = ""
ID_SEP = "/"

# _accounts is populated when get_swift_accounts() is called; which is
# guaranteed to be called before get_swift_containers and get_swift_objects
# since account plugin is parent and grandparent of container and object
_accounts = []

dateformat = "%a, %d %b %Y %H:%M:%S %Z"


def serialize_swift_account(account):
    metadocument = {k: account.get(k, None) for k, v in account.items()
                    if k.lower().startswith("x-account-meta")}
    account_fields = ('id', 'name')
    document = {f: account.get(f, None) for f in account_fields}

    document['domain_id'] = account.get('x-account-project-domain-id', None)
    if account.get('x-timestamp'):
        timestamp = float(account.get('x-timestamp'))
        document['created_at'] = \
            utils.isotime(datetime.datetime.fromtimestamp(timestamp))

    # lakshmiS: swift get_account() doesn't include update datetime field(?)
    if account.get('updated_at'):
        document['updated_at'] = account.get('updated_at')

    document.update(metadocument)
    return document


def serialize_swift_account_notification(account):
    account['name'] = account['project_name']
    account['id'] = account['account']
    account['x-account-project-domain-id'] = account['project_domain_id']
    # Note: updated_at is included in notification payload.
    # No need to map or transform the date and its format.
    return serialize_swift_account(account)


def serialize_swift_container(container):
    metadocument = {k: container.get(k, None) for k, v in
                    container.items()
                    if k.lower().startswith("x-container-meta")}
    container_fields = ('id',
                        'name',
                        'account',
                        'account_id',
                        'x-container-read'
                        )
    document = {f: container.get(f, None) for f in container_fields}
    if container.get('x-timestamp'):
        timestamp = float(container.get('x-timestamp'))
        document['created_at'] = \
            utils.isotime(datetime.datetime.fromtimestamp(timestamp))

    # (laskhmiS) get_container doesn't include last_modified field even
    # though the swift api documentation says it returns it. Include it
    # when it starts sending it.

    # Notifications for container sends this field as 'updated_at' instead
    # of 'last_modified'.
    if container.get('updated_at'):
        document['updated_at'] = container['updated_at']

    document.update(metadocument)
    return document


def serialize_swift_container_notification(container):
    # Account Id + container name. container['account'] from notification has
    # account id value.
    container['id'] = container['account'] + ID_SEP + container['container']
    container['name'] = container['container']
    container['account_id'] = container['account']
    container['account'] = container['project_name']
    return serialize_swift_container(container)


def serialize_swift_object(sobject):
    metadocument = {k: sobject.get(k, None) for k, v in sobject.items()
                    if k.lower().startswith("x-object-meta")}
    object_fields = ('id',
                     'name',
                     'account',
                     'account_id',
                     'container',
                     'container_id',
                     'etag'
                     )
    document = {f: sobject.get(f, None) for f in object_fields}
    document['content_type'] = sobject.get('content-type', None)
    document['content_length'] = sobject.get('content-length', None)

    if sobject.get('x-timestamp'):
        timestamp = float(sobject.get('x-timestamp'))
        document['created_at'] = \
            utils.isotime(datetime.datetime.fromtimestamp(timestamp))
    if sobject.get('last-modified'):
        updated_dt = datetime.datetime.strptime(
            sobject['last-modified'], dateformat)
        document['updated_at'] = utils.isotime(updated_dt)
    document.update(metadocument)
    return document


def serialize_swift_object_notification(sobj):
    # Account id + container name
    sobj['container_id'] = sobj['account'] + ID_SEP + sobj['container']
    # Account id + container name + object name
    sobj['id'] = sobj['container_id'] + ID_SEP + sobj['object']
    sobj['name'] = sobj['object']
    sobj['account_id'] = sobj['account']
    sobj['account'] = sobj['project_name']
    return serialize_swift_object(sobj)


@openstack_clients.clear_cached_swiftclient_on_unauthorized
def _get_storage_url_prefix():
    # Extracts swift proxy url after removing the default account id
    # from the service account. Later storage_url's will be constructed
    # for each account by appending the keystone tenant id.
    try:
        storage_url = openstack_clients.get_swiftclient().get_auth()[0]
        return storage_url[:storage_url.index(AUTH_PREFIX)] + AUTH_PREFIX
    except ValueError:
        LOG.error("reseller_prefix %s not found in keystone endpoint "
                  % AUTH_PREFIX)
        raise


def get_swift_accounts(auth_prefix):
    global AUTH_PREFIX
    # TODO(lakshmiS): Add support for SERVICE_ accounts
    AUTH_PREFIX = auth_prefix
    ks_client = openstack_clients.get_keystoneclient()
    for tenant in ks_client.projects.list():
        storage_url = _get_storage_url_prefix() + tenant.id
        sclient = openstack_clients.get_swiftclient_st(storage_url)
        # 0 index has account summary
        account = sclient.get_account()[0]
        account['name'] = tenant.name
        account['id'] = auth_prefix + tenant.id

        # store it for later usage in retrieving containers
        # and objects
        account_detail = {'id': account['id'],
                          'name': tenant.name,
                          'storage.url': storage_url}
        _accounts.append(account_detail)

        yield account


def get_swift_containers():
    for account in _accounts:
        sclient = openstack_clients.get_swiftclient_st(account['storage.url'])
        # 1 index has container list
        containers = sclient.get_account()[1]
        for container in containers:
            ctr, obj = sclient.get_container(container['name'])
            ctr['id'] = account['id'] + ID_SEP + container['name']
            ctr['name'] = container['name']
            ctr['account'] = account['name']
            ctr['account_id'] = account['id']
            yield ctr


def get_swift_objects():
    for account in _accounts:
        sclient = openstack_clients.get_swiftclient_st(account['storage.url'])
        # 1 index has container list
        containers = sclient.get_account()[1]
        for sctr in containers:
            ctr, obj = sclient.get_container(sctr['name'])
            for sobject in obj:
                sobj = sclient.head_object(sctr['name'], sobject['name'])
                sobj['account'] = account['name']
                sobj['account_id'] = account['id']
                sobj['container'] = sctr['name']
                sobj['container_id'] = account['id'] + ID_SEP + sctr['name']
                sobj['id'] = sobj['container_id'] + ID_SEP + sobject['name']
                sobj['name'] = sobject['name']
                yield sobj



# Copyright (c) 2016 Hewlett-Packard Enterprise Development L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from oslo_config import cfg
from searchlight import service_policies
import searchlight.tests.utils as test_utils


CONF = cfg.CONF


class TestServicePolicies(test_utils.BaseTestCase):
    def setUp(self):
        super(TestServicePolicies, self).setUp()

    def test_get_enforcers_abs_path(self):
        """Test service policy creation from cfg.policy_files"""

        nova_path = os.path.join(self.conf_dir, 'nova.json')
        with open(nova_path, 'w') as nova_policy_file:
            nova_policy_file.write('{"test": ""}')

        with mock.patch.object(CONF, 'service_policies') as mock_serv_pols:
            mock_serv_pols.service_policy_files = {'compute': nova_path}
            mock_serv_pols.service_policy_path = ''

            enforcers = service_policies._get_enforcers()
            self.assertEqual(['compute'], list(enforcers.keys()))
            self.assertEqual(['test'], list(enforcers['compute'].rules.keys()))

            # Check that get_enforcer_for_service works as expected
            self.assertIs(service_policies.get_enforcer_for_service('compute'),
                          enforcers['compute'])
            self.assertIsNone(service_policies.get_enforcer_for_service('aa'))

    def test_get_enforcers_base_path(self):
        """Test service policy creation with cfg.policy_path set"""
        nova_path = os.path.join(self.conf_dir, 'nova.json')
        with open(nova_path, 'w') as nova_policy_file:
            nova_policy_file.write('{"test": ""}')

        with mock.patch.object(CONF, 'service_policies') as mock_serv_pols:
            mock_serv_pols.service_policy_files = {'compute': 'nova.json'}
            mock_serv_pols.service_policy_path = self.conf_dir

            enforcers = service_policies._get_enforcers()
            self.assertEqual(['compute'], list(enforcers.keys()))
            self.assertEqual(['test'], list(enforcers['compute'].rules.keys()))

            # Check that get_enforcer_for_service works as expected
            self.assertIs(service_policies.get_enforcer_for_service('compute'),
                          enforcers['compute'])
            self.assertIsNone(service_policies.get_enforcer_for_service('aa'))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from searchlight.elasticsearch.plugins import base
from searchlight import pipeline


class MetadefHandler(base.NotificationBase):

    def __init__(self, *args, **kwargs):
        super(MetadefHandler, self).__init__(*args, **kwargs)
        self.namespace_delete_keys = ['deleted_at', 'deleted',
                                      'namespace_old']
        self.property_delete_keys = ['deleted', 'deleted_at',
                                     'name_old', 'namespace']

    @classmethod
    def _get_notification_exchanges(cls):
        return ['glance']

    def get_event_handlers(self):
        return {
            "metadef_namespace.create": self.create_ns,
            "metadef_namespace.update": self.update_ns,
            "metadef_namespace.delete": self.delete_ns,
            "metadef_object.create": self.create_obj,
            "metadef_object.update": self.update_obj,
            "metadef_object.delete": self.delete_obj,
            "metadef_property.create": self.create_prop,
            "metadef_property.update": self.update_prop,
            "metadef_property.delete": self.delete_prop,
            "metadef_resource_type.create": self.create_rs,
            "metadef_resource_type.delete": self.delete_rs,
            "metadef_tag.create": self.create_tag,
            "metadef_tag.update": self.update_tag,
            "metadef_tag.delete": self.delete_tag,
            "metadef_namespace.delete_properties": self.delete_props,
            "metadef_namespace.delete_objects": self.delete_objects,
            "metadef_namespace.delete_tags": self.delete_tags
        }

    def get_log_fields(self, event_type, payload):
        return ('namespace', payload.get('namespace')),

    def create_ns(self, event_type, payload, timestamp):
        namespace = self.format_namespace(payload)
        self.index_helper.save_document(
            namespace,
            version=self.get_version(namespace, timestamp))
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  namespace
                                  )

    def update_ns(self, event_type, payload, timestamp):
        # Update operation in es doesn't support external version,
        # so we have to manually update the doc and reindex it.
        namespace_es = self.index_helper.get_document(
            payload['namespace_old'], for_admin=True)
        namespace = self.format_namespace(payload)
        namespace_es['_source'].update(namespace)
        self.index_helper.save_document(
            namespace_es['_source'],
            version=self.get_version(namespace, timestamp))
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  namespace_es['_source'])

    def delete_ns(self, event_type, payload, timestamp):
        id = payload['namespace']
        self.index_helper.delete_document({'_id': id})
        return pipeline.DeleteItem(self.index_helper.plugin,
                                   event_type,
                                   payload,
                                   id
                                   )

    def create_obj(self, event_type, payload, timestamp):
        id = payload['namespace']
        object = self.format_object(payload)
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.create_entity(preexisting['_source'], 'objects', object)
        self.index_helper.save_document(
            preexisting['_source'], preexisting['_version'] + 1
        )
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def update_obj(self, event_type, payload, timestamp):
        id = payload['namespace']
        object = self.format_object(payload)
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.update_entity(
            preexisting['_source'], "objects",
            payload['name_old'],
            object,
            "name")
        self.index_helper.save_document(
            preexisting['_source'], preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def delete_obj(self, event_type, payload, timestamp):
        id = payload['namespace']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.delete_entity(
            preexisting['_source'], "objects", payload['name'], "name")
        self.index_helper.save_document(
            preexisting['_source'], preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def create_prop(self, event_type, payload, timestamp):
        id = payload['namespace']
        property = self.format_property(payload)
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.create_entity(preexisting['_source'], 'properties', property)
        self.index_helper.save_document(
            preexisting['_source'], preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def update_prop(self, event_type, payload, timestamp):
        id = payload['namespace']
        property = self.format_property(payload)
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.update_entity(
            preexisting['_source'],
            "properties",
            payload['name_old'],
            property,
            "name")
        self.index_helper.save_document(
            preexisting['_source'], preexisting['_version'] + 1
        )
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def delete_prop(self, event_type, payload, timestamp):
        id = payload['namespace']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.delete_entity(
            preexisting['_source'], "properties", payload['name'], "name")
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def create_rs(self, event_type, payload, timestamp):
        id = payload['namespace']
        resource_type = {}
        resource_type['name'] = payload['name']
        if payload['prefix']:
            resource_type['prefix'] = payload['prefix']
        if payload['properties_target']:
            resource_type['properties_target'] = payload['properties_target']

        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.create_entity(
            preexisting['_source'], "resource_types", resource_type)
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def delete_rs(self, event_type, payload, timestamp):
        id = payload['namespace']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.delete_entity(
            preexisting['_source'], "resource_types", payload['name'], "name")
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def create_tag(self, event_type, payload, timestamp):
        id = payload['namespace']
        tag = {}
        tag['name'] = payload['name']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.create_entity(preexisting['_source'], "tags", tag)
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def update_tag(self, event_type, payload, timestamp):
        id = payload['namespace']
        tag = {}
        tag['name'] = payload['name']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.update_entity(
            preexisting['_source'], "tags", tag, payload['name_old'], "name")
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def delete_tag(self, event_type, payload, timestamp):
        id = payload['namespace']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        self.delete_entity(
            preexisting['_source'], "tags", payload['name'], "name")
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def delete_props(self, event_type, payload, timestamp):
        return self.delete_field(
            event_type,
            payload,
            timestamp,
            'properties'
        )

    def delete_objects(self, event_type, payload, timestamp):
        return self.delete_field(
            event_type,
            payload,
            timestamp,
            'objects'
        )

    def delete_tags(self, event_type, payload, timestamp):
        return self.delete_field(
            event_type,
            payload,
            timestamp,
            'tags'
        )

    def delete_field(self, event_type, payload, timestamp, field):
        id = payload['namespace']
        preexisting = self.index_helper.get_document(id, for_admin=True)
        preexisting['_source'].pop(field, None)
        self.index_helper.save_document(preexisting['_source'],
                                        preexisting['_version'] + 1)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  preexisting['_source'])

    def format_namespace(self, payload):
        for key in self.namespace_delete_keys:
            if key in payload.keys():
                del payload[key]
        payload['id'] = payload['namespace']
        if 'display_name' in payload and payload['display_name']:
            payload['name'] = payload['display_name']
        else:
            payload['name'] = payload['namespace']
        return payload

    def format_object(self, payload):
        formatted_object = {}
        formatted_object['name'] = payload['name']
        formatted_object['description'] = payload['description']
        if payload['required']:
            formatted_object['required'] = payload['required']
        formatted_object['properties'] = []
        for property in payload['properties']:
            formatted_property = self.format_property(property)
            formatted_object['properties'].append(formatted_property)
        return formatted_object

    def format_property(self, payload):
        prop_data = {}
        for key, value in payload.items():
            if key not in self.property_delete_keys and value:
                prop_data[key] = value
        return prop_data

    def create_entity(self, doc, entity_name, entity_data):
        entity_list = doc.setdefault(entity_name, [])
        entity_list.append(entity_data)
        return doc

    def update_entity(self, doc, entity_name, entity_id, entity_data,
                      field_name):
        self.delete_entity(doc, entity_name, entity_id, field_name)
        self.create_entity(doc, entity_name, entity_data)
        return doc

    def delete_entity(self, doc, entity_name, entity_id, field_name):
        match_entity = None
        for index, item in enumerate(doc.get(entity_name, [])):
            if item[field_name] == entity_id:
                match_entity = index
                break
        if match_entity is not None:
            doc[entity_name].pop(match_entity)
        return doc



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cinderclient.exceptions
from oslo_log import log as logging

from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.cinder import serialize_cinder_snapshot
from searchlight import pipeline

LOG = logging.getLogger(__name__)


class SnapshotHandler(base.NotificationBase):
    """Handles cinder snapshot notifications. These can come as a result of
    a user action (like a create, delete, metadata edit etc) or as a result of
    periodic auditing notifications cinder sends
    """

    @classmethod
    def _get_notification_exchanges(cls):
        # Unlike most services cinder doesn't override the exchange name from
        # the oslo.messaging default
        return ['openstack']

    def get_event_handlers(self):
        return {
            'snapshot.update.end': self.create_or_update,
            'snapshot.create.end': self.create_or_update,
            'snapshot.delete.end': self.delete,
        }

    def get_log_fields(self, event_type, payload):
        return (
            ('id', payload.get('snapshot_id')),
            ('volume_id', payload.get('volume_id'))
        )

    def create_or_update(self, event_type, payload, timestamp):
        snapshot_id = payload['snapshot_id']
        LOG.debug("Updating cinder snapshot information for %s", snapshot_id)
        try:
            snapshot_payload = serialize_cinder_snapshot(snapshot_id)
            version = self.get_version(snapshot_payload, timestamp)
            self.index_helper.save_document(snapshot_payload, version=version)
            return pipeline.IndexItem(
                self.index_helper.plugin,
                event_type,
                payload,
                snapshot_payload)
        except cinderclient.exceptions.NotFound:
            LOG.warning("Snapshot %s not found; deleting" % snapshot_id)
            self.delete(payload, timestamp)

    def delete(self, event_type, payload, timestamp):
        snapshot_id = payload['snapshot_id']
        volume_id = payload['volume_id']
        LOG.debug("Deleting cinder snapshot information for %s", snapshot_id)
        if not snapshot_id:
            return

        try:
            self.index_helper.delete_document({'_id': snapshot_id,
                                               '_parent': volume_id})
            return pipeline.DeleteItem(
                self.index_helper.plugin, event_type, payload, snapshot_id)
        except Exception as exc:
            LOG.error(
                'Error deleting snapshot %(snapshot_id)s '
                'from index. Error: %(exc)s' %
                {'snapshot_id': snapshot_id, 'exc': exc})



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from searchlight.listener import NotificationEndpoint
from searchlight.pipeline import PipelineManager
from searchlight.tests import functional
from searchlight.tests.functional import test_api
from searchlight.tests.functional import test_listener
from searchlight.tests import utils


SERVER_ID = u"ae81266c-a292-4e48-bf81-a3894bd62f4a"
TENANT_ID1 = u"1816a16093df465dbc609cf638422a05"
TENANT_ID2 = u"75c31cdaa3604b76b7e279de50aec9f0"
USER_ID = u"3f7600957f0a444c98928d79debb72fb"
VOLUME_ID1 = u"f54b1d2f-4514-4688-8da7-4bc66767b6be"
VOLUME_ID2 = u"fe7a158a-7f6d-461f-84df-83d394f7c0f6"
VOLUME_ID3 = u"faaad6fe-9351-4313-bce9-881b476d5751"
SNAP_ID1 = u"6d2643b5-d579-417f-b9b9-969ce94dc134"
SNAP_ID2 = u"56f91625-2fc5-4e4a-8284-4bfd3b3a1914"


volume_manager = 'cinderclient.v2.volumes.VolumeManager'
snapshot_manager = 'cinderclient.v2.volume_snapshots.SnapshotManager'


class TestCinderPlugins(functional.FunctionalTest):
    def setUp(self):
        super(TestCinderPlugins, self).setUp()
        self.volume_plugin = self.initialized_plugins['OS::Cinder::Volume']
        self.snapshot_plugin = self.initialized_plugins['OS::Cinder::Snapshot']

        self.volume_objects = self._load_fixture_data('load/volumes.json')
        self.snapshot_objects = self._load_fixture_data('load/snapshots.json')

    def _index_data(self):
        """Moving this here because failures in setUp result in the API server
        not getting torn down properly.
        """
        def fake_volume_get(volume_id):
            vol = list(filter(lambda v: v["id"] == volume_id,
                       self.volume_objects))[0]
            return utils.DictObj(**vol)

        def fake_snapshot_get(snapshot_id):
            snap = list(filter(lambda v: v["id"] == snapshot_id,
                        self.snapshot_objects))[0]
            return utils.DictObj(**snap)

        self._index(self.snapshot_plugin,
                    [utils.DictObj(**snap) for snap in self.snapshot_objects])

        self._index(self.volume_plugin,
                    [utils.DictObj(**vol) for vol in self.volume_objects])

    def test_query_attachments(self):
        self._index_data()
        query = {
            "type": "OS::Cinder::Volume",
            "query": {
                "nested": {
                    "path": "attachments",
                    "query": {
                        "bool": {
                            "must": [
                                {"term": {"attachments.device": "/dev/vdb"}},
                                {"term": {"attachments.server_id": SERVER_ID}}
                            ]
                        }
                    }
                }
            }
        }
        response, json_content = self._search_request(query,
                                                      TENANT_ID1)
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, json_content['hits']['total'])

    def test_snapshots_of_volume(self):
        self._index_data()
        query = {
            "type": "OS::Cinder::Snapshot",
            "query": {
                "has_parent": {
                    "parent_type": "OS::Cinder::Volume",
                    "query": {
                        "term": {"id": "fe7a158a-7f6d-461f-84df-83d394f7c0f6"}
                    }
                }
            }
        }
        response, json_content = self._search_request(query,
                                                      TENANT_ID1)
        self.assertEqual(200, response.status_code)
        self.assertEqual(1, json_content['hits']['total'])
        self.assertEqual(SNAP_ID1,
                         json_content['hits']['hits'][0]['_source']['id'])

    def test_rbac(self):
        self._index_data()
        response, json_content = self._search_request(test_api.MATCH_ALL,
                                                      TENANT_ID1)
        self.assertEqual(200, response.status_code)
        hits = [hit['_source'] for hit in json_content['hits']['hits']]
        self.assertEqual(set([VOLUME_ID1, VOLUME_ID2, SNAP_ID1]),
                         set(h['id'] for h in hits))


class TestCinderNotifications(test_listener.TestSearchListenerBase):
    """Cinder has to go to the API currently so we can't really test the
    search results from notifications, just that there was an attempt to
    retrieve and store the results
    """
    def setUp(self):
        super(TestCinderNotifications, self).setUp()
        self.volume_plugin = self.initialized_plugins['OS::Cinder::Volume']
        self.snapshot_plugin = self.initialized_plugins['OS::Cinder::Snapshot']

        self.volume_events = self._load_fixture_data('events/volumes.json')
        self.snapshot_events = self._load_fixture_data('events/snapshots.json')

        notification_plugins = {
            plugin.document_type: utils.StevedoreMock(plugin)
            for plugin in (self.volume_plugin, self.snapshot_plugin)}
        self.notification_endpoint = NotificationEndpoint(
            notification_plugins,
            PipelineManager(notification_plugins)
        )

        self.index_alias = self.volume_plugin.alias_name_listener

        volume = {
            'id': VOLUME_ID1,
            'user_id': USER_ID,
            'created_at': '2016-03-07T16:51:09.000000',
            'updated_at': '2016-03-07T16:51:09.000000'
        }
        self.volume_fixture = utils.DictObj(**volume)

    @mock.patch(volume_manager + '.get')
    def test_volume_cud(self, mock_volume_get):
        mock_volume_get.return_value = self.volume_fixture

        vol_create = self.volume_events['volume.create.end']
        self._send_event_to_listener(vol_create, self.index_alias)

        vol_update = self.volume_events['volume.update.end']
        self._send_event_to_listener(vol_update, self.index_alias)

        vol_delete = self.volume_events['volume.delete.end']
        self._send_event_to_listener(vol_delete, self.index_alias)

        mock_volume_get.assert_has_calls([
            mock.call(VOLUME_ID1), mock.call(VOLUME_ID1)
        ])

    @mock.patch(volume_manager + '.get')
    def test_volume_attach(self, mock_volume_get):
        mock_volume_get.return_value = self.volume_fixture

        vol_attach = self.volume_events['volume.attach.end']
        self._send_event_to_listener(vol_attach, self.index_alias)

        vol_detach = self.volume_events['volume.detach.end']
        self._send_event_to_listener(vol_detach, self.index_alias)

        mock_volume_get.assert_has_calls([
            mock.call(VOLUME_ID1), mock.call(VOLUME_ID1)
        ])

    @mock.patch(volume_manager + '.get')
    def test_volume_retype(self, mock_volume_get):
        mock_volume = mock.Mock(id=VOLUME_ID1, user_id=USER_ID,
                                created_at='2016-03-07T16:51:09.000000',
                                updated_at='2016-03-07T16:51:09.000000')
        mock_volume_get.return_value = mock_volume

        vol_retype = self.volume_events['volume.retype']
        self._send_event_to_listener(vol_retype, self.index_alias)

        mock_volume_get.assert_called_with(VOLUME_ID1)

    @mock.patch(snapshot_manager + '.get')
    def test_snapshot_create_delete(self, mock_snapshot_get):
        mock_snap = mock.Mock(id=SNAP_ID2,
                              created_at='2016-03-07T16:51:09.000000',
                              updated_at='2016-03-07T16:51:09.000000')
        mock_snapshot_get.return_value = mock_snap

        snapshot_create = self.snapshot_events['snapshot.create.end']
        self._send_event_to_listener(snapshot_create, self.index_alias)

        snapshot_delete = self.snapshot_events['snapshot.delete.end']
        self._send_event_to_listener(snapshot_delete, self.index_alias)

        mock_snapshot_get.assert_has_calls([mock.call(SNAP_ID2)])



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from elasticsearch import exceptions
from oslo_log import log as logging

from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins import designate
from searchlight import pipeline


LOG = logging.getLogger(__name__)


class ZoneHandler(base.NotificationBase):
    def __init__(self, *args, **kwargs):
        self.recordset_helper = kwargs.pop('recordset_helper')
        super(ZoneHandler, self).__init__(*args, **kwargs)
        self.domain_delete_keys = ['deleted_at', 'deleted',
                                   'attributes', 'recordsets']

    @classmethod
    def _get_notification_exchanges(cls):
        return ['designate']

    def get_event_handlers(self):

        # domain(v1) and zone(v2) are same except for the name.
        # To keep it backward compatible, designate sends two sets of
        # notification events (dns.zone.xxxx and dns.domain.xxxx) for
        # every domain or zone action within v1 or v2 api.
        # So we ignore all dns.domain.xxxx events.
        return {
            "dns.zone.create": self.create_or_update,
            "dns.zone.update": self.create_or_update,
            "dns.zone.delete": self.delete
        }

    def _serialize(self, payload):
        for key in self.domain_delete_keys:
            if key in payload:
                del payload[key]

        if 'masters' in payload:
            payload['masters'] = ["%(host)s:%(port)s" for i in
                                  payload["masters"]]
        payload['project_id'] = payload.pop('tenant_id')
        if not payload['updated_at'] and payload['created_at']:
            payload['updated_at'] = payload['created_at']

        return payload

    def process(self, ctxt, publisher_id, event_type, payload, metadata):
        if (event_type == 'dns.zone.update' and
                payload.get('status') == 'DELETED'):
            LOG.debug("Ignoring update notification for Domain with DELETED "
                      "status; event will be processed on delete event")
            return None

        items = super(ZoneHandler, self).process(
            ctxt, publisher_id, event_type, payload, metadata)
        try:
            # NOTE: So if this is an initial zone we need to index the SOA / NS
            # records it will have. Let's do this when receiving the create
            # event.
            if event_type == 'dns.zone.create':
                if not items:
                    LOG.warning("Not writing initial recordsets; exception "
                                "occurred during zone indexing")
                    return None

                recordsets = designate._get_recordsets(payload['id'])
                serialized_recordsets = []
                recordset_versions = []
                for rs in recordsets:
                    rs = designate._serialize_recordset(rs)

                    # So project ID isn't provided in the recordset api.
                    rs['project_id'] = payload['project_id']

                    serialized_recordsets.append(rs)

                    # Use the timestamp from *this* notification but the
                    # updated_at from each recordset (which empirically appears
                    # to be the same for all initial recordsets)
                    recordset_versions.append(
                        self.get_version(rs, metadata['timestamp']))
                    items.append(
                        pipeline.IndexItem(self.recordset_helper.plugin,
                                           event_type,
                                           payload,
                                           rs
                                           )
                    )
                self.recordset_helper.save_documents(
                    serialized_recordsets, versions=recordset_versions)
                return items
        except Exception as e:
            LOG.exception(e)

    def create_or_update(self, event_type, payload, timestamp):
        zone_payload = self._serialize(payload)
        self.index_helper.save_document(
            zone_payload,
            version=self.get_version(zone_payload, timestamp))
        return pipeline.IndexItem(
            self.index_helper.plugin,
            event_type,
            payload,
            zone_payload)

    def delete(self, event_type, payload, timestamp):
        zone_id = payload['id']
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        delete_recordset = self.recordset_helper.delete_documents_with_parent(
            zone_id,
            version=version)
        items = [pipeline.DeleteItem(
            self.recordset_helper.plugin,
            event_type,
            payload,
            rs['_id']) for rs in delete_recordset]

        try:
            self.index_helper.delete_document(
                {'_id': zone_id, '_version': version})
            items.append(pipeline.DeleteItem(self.index_helper.plugin,
                                             event_type,
                                             payload,
                                             zone_id))
        except exceptions.NotFoundError:
            msg = "Zone %s not found when deleting"
            LOG.error(msg, zone_id)
        return items


class RecordSetHandler(base.NotificationBase):
    # TODO(sjmc7): see note above
    DOCUMENT_TYPE = "OS::Designate::RecordSet"

    def __init__(self, *args, **kwargs):
        super(RecordSetHandler, self).__init__(*args, **kwargs)
        self.record_delete_keys = ['deleted_at', 'deleted',
                                   'attributes']

    @classmethod
    def _get_notification_exchanges(cls):
        return ['designate']

    def get_event_handlers(self):
        return {
            "dns.recordset.create": self.create_or_update_recordset,
            "dns.recordset.update": self.create_or_update_recordset,
            "dns.recordset.delete": self.delete_recordset,
        }

    def get_log_fields(self, event_type, payload):
        return (
            ('id', payload.get('id')),
            ('zone_id', payload.get('zone_id'))
        )

    def create_or_update_recordset(self, event_type, payload, timestamp):
        rs_payload = self._serialize(payload)
        self.index_helper.save_document(
            rs_payload,
            version=self.get_version(rs_payload, timestamp))
        return pipeline.IndexItem(
            self.index_helper.plugin,
            event_type,
            payload,
            rs_payload)

    def _serialize(self, obj):
        obj['project_id'] = obj.pop('tenant_id')
        obj['records'] = [i['data'] for i in obj['records']]
        if not obj['updated_at'] and obj['created_at']:
            obj['updated_at'] = obj['created_at']
        return obj

    def delete_recordset(self, event_type, payload, timestamp):
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        self.index_helper.delete_document(
            {'_id': payload['id'], '_version': version,
             '_parent': payload['zone_id']})
        return pipeline.DeleteItem(self.index_helper.plugin,
                                   event_type,
                                   payload,
                                   payload['id'])



#!/usr/bin/env python

# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from oslo_config import cfg
import oslo_messaging
from oslo_serialization import jsonutils
import six.moves.urllib.parse as urlparse
import sys
import time

topic = 'notifications'
password = os.environ.get('RABBIT_PASSWORD', os.environ.get('OS_PASSWORD'))
host = urlparse.urlparse(os.environ.get('OS_AUTH_URL')).hostname
username = os.environ.get('RABBIT_USER', 'stackrabbit')


class EP(object):
    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        all_locals = locals()
        all_locals.pop('self')
        print(jsonutils.dumps(all_locals))

    def error(self, ctxt, publisher_id, event_type, payload, metadata):
        all_locals = locals()
        all_locals.pop('self')
        print(jsonutils.dumps(all_locals))


def main():
    if len(sys.argv) < 2:
        print("Supply an exchange")
        sys.exit(0)

    exchange = sys.argv[1]
    pool = sys.argv[2] if len(sys.argv) > 2 else None

    transport = oslo_messaging.get_notification_transport(
        cfg.CONF,
        url='rabbit://%s:%s@%s' % (username, password, host))
    targets = [oslo_messaging.Target(topic=topic, exchange=exchange)]
    endpoints = [EP()]
    oslo_listener = oslo_messaging.get_notification_listener(
        transport, targets, endpoints, pool=pool, executor='threading')
    try:
        print("Started")
        oslo_listener.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Stopping")
        oslo_listener.stop()
        oslo_listener.wait()


if __name__ == '__main__':
    main()



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from searchlight.elasticsearch.plugins import base
from searchlight import listener
import searchlight.tests.utils as test_utils


class PretendNotificationHandler(base.NotificationBase):
    def __init__(self, *args, **kwargs):
        index_helper = mock.Mock()
        options = mock.Mock()
        super(PretendNotificationHandler, self).__init__(index_helper, options,
                                                         *args, **kwargs)
        self.event1_called_with = []

    @classmethod
    def _get_notification_exchanges(cls):
        return []

    def get_event_handlers(self):
        return {
            'test.create.end': self.event1
        }

    def event1(self, event_type, payload, timestamp):
        self.event1_called_with.append((event_type, payload, timestamp))


class StevedorePlugin(object):
    def __init__(self, name, plugin):
        self._plugin = plugin
        self._name = name

    @property
    def name(self):
        return self._name

    @property
    def obj(self):
        return self._plugin


class TestListener(test_utils.BaseTestCase):
    def test_multiple_event_handlers(self):
        handler1 = PretendNotificationHandler()
        plugin1 = mock.Mock()
        plugin1.get_notification_handler.return_value = handler1

        handler2 = PretendNotificationHandler()
        plugin2 = mock.Mock()
        plugin2.get_notification_handler.return_value = handler2

        plugins = {'plugin1': StevedorePlugin('plugin1', plugin1),
                   'plugin2': StevedorePlugin('plugin1', plugin2)}
        notification_ep = listener.NotificationEndpoint(plugins, mock.Mock())

        self.assertEqual(['test.create.end'],
                         list(notification_ep.notification_target_map.keys()))
        self.assertEqual(
            set([plugin1, plugin2]),
            set(notification_ep.notification_target_map['test.create.end']))

        notification_ep.info({}, None, 'test.create.end',
                             {'test': 'payload'}, {'timestamp': 1234})

        self.assertEqual([('test.create.end', {'test': 'payload'}, 1234)],
                         handler1.event1_called_with)
        self.assertEqual([('test.create.end', {'test': 'payload'}, 1234)],
                         handler2.event1_called_with)


@mock.patch.object(listener.LOG, 'info')
class TestEndpointPriorities(test_utils.BaseTestCase):
    def test_priorities(self, log_info_mock):
        handler = PretendNotificationHandler()
        plugin = mock.Mock()
        plugin.get_notification_handler.return_value = handler

        plugins = {'plugin': StevedorePlugin('plugin', plugin)}
        notification_ep = listener.NotificationEndpoint(plugins, mock.Mock())
        fake_metadata = {'timestamp': 'fake'}
        notification_ep.info({}, 'fake', 'test.create.end', {}, fake_metadata)
        self.assertEqual('INFO', log_info_mock.call_args[0][1]['priority'])
        notification_ep.error({}, 'fake', 'test.create.end', {}, fake_metadata)
        self.assertEqual('ERROR', log_info_mock.call_args[0][1]['priority'])



# Copyright (c) 2015-2016 Hewlett-Packard Enterprise Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from elasticsearch import exceptions as es_exc
from elasticsearch import helpers
import logging
import oslo_utils
import six

from oslo_config import cfg
from oslo_utils import encodeutils

from searchlight.common import exception as sl_exc
from searchlight.common import utils
from searchlight.context import RequestContext
import searchlight.elasticsearch


# Format for datetime when creating a unique index.
FORMAT = "%Y_%m_%d_%H_%M_%S"

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

VERSION_CONFLICT_MSG = 'version_conflict_engine_exception'


def get_now_str():
    """Wrapping this to make testing easier (mocking utcnow's troublesome)
    and keep it in one place in case oslo changes
    """
    return utils.isotime(oslo_utils.timeutils.utcnow())


def timestamp_to_isotime(timestamp):
    """Take a rabbitmq-style timestamp (2016-03-14 16:49:23.749458) and convert
    to an ISO8601 timestamp.
    """
    fmt = '%Y-%m-%d %H:%M:%S.%f'
    parsed_time = datetime.datetime.strptime(timestamp, fmt)
    return utils.isotime(parsed_time)


def helper_reindex(client, source_index, target_index, query=None,
                   target_client=None, chunk_size=500, scroll='5m',
                   scan_kwargs={}, bulk_kwargs={}):
    """We have lovingly copied the entire helpers.reindex() method from the
       elasticsearch Python client. The original version lives here:
           https://github.com/elastic/elasticsearch-py
       In the file:
           elasticsearch/helpers/__init__.py
       This file is licensed under the Apache License.

       We need our own version (haha) of reindex to handle external versioning
       within a document. The current implmenentation of helpers.reindex does
       not provide this support. Since there is no way to tell helpers.bulk()
       that an external version is being used, we will need to modify each
       document in the generator. For future maintainablilty, modifications
       to the original method will be preceded with a comment "CHANGED:".
       Minor tweaks made for PEP8 conformance excepted.
    """

    """
    Reindex all documents from one index that satisfy a given query
    to another, potentially (if `target_client` is specified) on a different
    cluster. If you don't specify the query you will reindex all the documents.

    .. note::

        This helper doesn't transfer mappings, just the data.

    :arg client: instance of :class:`~elasticsearch.Elasticsearch` to use (for
        read if `target_client` is specified as well)
    :arg source_index: index (or list of indices) to read documents from
    :arg target_index: name of the index in the target cluster to populate
    :arg query: body for the :meth:`~elasticsearch.Elasticsearch.search` api
    :arg target_client: optional, is specified will be used for writing (thus
        enabling reindex between clusters)
    :arg chunk_size: number of docs in one chunk sent to es (default: 500)
    :arg scroll: Specify how long a consistent view of the index should be
        maintained for scrolled search
    :arg scan_kwargs: additional kwargs to be passed to
        :func:`~elasticsearch.helpers.scan`
    :arg bulk_kwargs: additional kwargs to be passed to
        :func:`~elasticsearch.helpers.bulk`
    """
    target_client = client if target_client is None else target_client

    docs = helpers.scan(client, query=query, index=source_index,
                        scroll=scroll,
                        fields=('_source', '_parent', '_routing',
                                '_timestamp'), **scan_kwargs)

    def _change_doc_index(hits, index):
        for h in hits:
            h['_index'] = index
            # CHANGED: Allow for external versions to be indexed.
            h['_version_type'] = "external"
            if 'fields' in h:
                h.update(h.pop('fields'))
            yield h

    kwargs = {
        'stats_only': True,
    }
    kwargs.update(bulk_kwargs)
    return helpers.bulk(target_client, _change_doc_index(docs, target_index),
                        chunk_size=chunk_size, **kwargs)


def reindex(src_index, dst_index, type_list, chunk_size=None, time=None):
    """Reindex a set of indexes internally within ElasticSearch. All of the
       documents under the types that live in "type_list" under the index
       "src_index" will be copied into the documents under the same types
       in the index "dst_index". In other words, a perfect re-index!
       Instead of using the plugin API and consuming bandwidth to perform
       the re-index we will allow ElasticSearch to do some heavy lifting for
       us. Under the covers we are combining scan/scroll with bulk operations
       to do this re-indexing as efficient as possible.
    """
    es_engine = searchlight.elasticsearch.get_api()

    # Create a Query DSL string to access all documents within the specified
    # document types. We will filter on the "_type" field in this index. Since
    # there are multiple docuent types, we will need to use the "terms" filter.
    # All of the document types will be added to the list for "_type". We need
    # to enable version to allow the search to return the version field. This
    # will be used by the reindexer.
    body = {"version": "true",
            "query": {"bool": {"filter": {"terms": {"_type": type_list}}}}}
    # Debug: Show all documents that ES will re-index.
    # LOG.debug(es_engine.search(index=src_index, body=body, size=500))
    helper_reindex(client=es_engine, source_index=src_index,
                   target_index=dst_index, query=body)


def create_new_index(group):
    """Create a new index for a specific Resource Type Group. Upon
       exit of this method, the index is still not ready to be used.
       The index still needs to have the settings/mappings set for
       each plugin (Document Type).
    """
    es_engine = searchlight.elasticsearch.get_api()

    kwargs = {}
    index_settings = _get_index_settings_from_config()
    if index_settings:
        kwargs = {'body': {'index': index_settings}}

    index_name = None
    while not index_name:
        # Use utcnow() to ensure that the name is unique.
        now = oslo_utils.timeutils.utcnow()
        index_name = (group + '-' + now.strftime(FORMAT))
        try:
            es_engine.indices.create(index=index_name, **kwargs)
        except es_exc.TransportError as e:
            if (e.error.startswith("IndexAlreadyExistsException") or
                    e.error.startswith("index_already_exists_exception")):
                # This index already exists! Try again.
                index_name = None
            else:
                raise

    return index_name


def _get_index_settings_from_config():
    index_settings = {}
    if CONF.elasticsearch.index_gc_deletes is not None:
        index_settings['gc_deletes'] = CONF.elasticsearch.index_gc_deletes

    for setting, value in CONF.elasticsearch.index_settings.items():
        if setting.startswith('index.'):
            setting = setting[len('index_'):]

        index_settings[setting] = value

    return index_settings


def add_extra_mappings(index_name, doc_type_info):
    """Add mappings for the specified doc_types if they already do not
       exist in the index. This is to work around a "feature" in Elasticsearch.
       Specifying a doc_type in a query that spans multiple indices will fail
       if that doc_type is not present in all indices being queried. This type
       of query is used in Searchlight for the Resource Types that have a
       parent-child relationship (due to our RBAC model).
       The parameter doc_type_info is a list of tuples. Each entry contains
       a document type and whether this document type has a parent-child
       relationship.
    """
    es_engine = searchlight.elasticsearch.get_api()

    # Create an empty mapping that cannot be used. We need to make sure it
    # has a "_parent" field for the "has_parent" queries. Make sure that the
    # "_parent" field is not used anywhere else.
    for doc_type, has_parent in doc_type_info:
        body = {'dynamic': 'strict',
                'properties': {
                }}
        if has_parent:
            body['_parent'] = {'type': 'never_used_parent'}
        if not es_engine.indices.exists_type(index=index_name,
                                             doc_type=doc_type):
            es_engine.indices.put_mapping(index=index_name, doc_type=doc_type,
                                          body=body)


def get_index_refresh_interval(index_name):
    """Get the refresh_interval of a given index, if refresh_interval isn't
       set, return default 1s.
    """

    es_engine = searchlight.elasticsearch.get_api()
    try:
        result = es_engine.indices.get_settings(index_name,
                                                'index.refresh_interval')
    except Exception as e:
        # If we cannot get index setting, something must be wrong,
        # no need to continue, log the error message and raise.
        LOG.error(encodeutils.exception_to_unicode(e))
        raise

    if result:
        return result[index_name]['settings']['index']['refresh_interval']
    else:
        return '1s'


def set_index_refresh_interval(index_name, refresh_interval):
    """Set refresh_interval of a given index, basically it is used in the
       reindexing phase. By setting refresh_interval to -1 we disable the
       refresh of offline index to gain a performance boost for the bulk
       updates. After reindexing is done, we will restore refresh_interval
       and put the index online.
    """

    es_engine = searchlight.elasticsearch.get_api()

    body = {
        'index': {
            'refresh_interval': refresh_interval
        }
    }

    try:
        es_engine.indices.put_settings(body, index_name)
    except Exception as e:
        LOG.error(encodeutils.exception_to_unicode(e))
        raise


def refresh_index(index_name):
    """Do a refresh on a given index"""

    es_engine = searchlight.elasticsearch.get_api()

    try:
        es_engine.indices.refresh(index_name)
    except Exception as e:
        LOG.error(encodeutils.exception_to_unicode(e))
        raise


def setup_alias(index_name, alias_search, alias_listener):
    """Create all needed aliases. Each Resource Type Group will have two
       aliases. Each alias will point to the same internal index. As a
       reminder, all ES CRUD operations should go to the listener alias.
       All ES Query operations should go to the search alias.
    """
    es_engine = searchlight.elasticsearch.get_api()

    if not es_engine.indices.exists_alias(name=alias_search):
        # Search alias does not exist, create it and continue.
        try:
            es_engine.indices.put_alias(index=index_name,
                                        name=alias_search)
        except Exception as e:
            LOG.error(encodeutils.exception_to_unicode(e))
            es_engine.indices.delete(index=index_name)
            raise

    if not es_engine.indices.exists_alias(name=alias_listener):
        # Listener alias does not exist, create it and return.
        try:
            es_engine.indices.put_alias(index=index_name,
                                        name=alias_listener)
        except Exception as e:
            LOG.error(encodeutils.exception_to_unicode(e))
            es_engine.indices.delete(index=index_name)
            raise
        return

    # Listener alias exists, add the new index to it.
    body = {
        'actions': [
            {'add': {
                'index': index_name,
                'alias': alias_listener}}
        ]
    }
    try:
        es_engine.indices.update_aliases(body=body)
    except Exception as e:
        LOG.error(encodeutils.exception_to_unicode(e))
        es_engine.indices.delete(index=index_name, ignore=404)
        raise


def alias_search_update(alias_search, index_name):
    """Replace the current index with the specified index in the
       search alias. To avoid a race condition we will need to
       perform the remove and add automatically (within ElasticSearch).
    """
    if not index_name:
        return None

    es_engine = searchlight.elasticsearch.get_api()

    body = {
        'actions': [
            {'add': {
                'index': index_name,
                'alias': alias_search}}
        ]
    }
    old_index = None
    try:
        current_indices = es_engine.indices.get_alias(name=alias_search)
        # Grab first (and only) index in the list from get_alias().
        old_index = list(current_indices.keys())[0]
        if old_index == index_name:
            return None
        body['actions'].insert(0,
                               {'remove': {
                                   'index': old_index,
                                   'alias': alias_search}})

    except es_exc.NotFoundError:
        # Alias doesn't exist. Strange. Nothing to remove, only an add.
        old_index = None

    try:
        es_engine.indices.update_aliases(body)
    except Exception as e:
        # The alias probably still refers to the index. Log the error and stop.
        # Continuing may result in the index being deleted and catastrophic
        # failure later.
        LOG.error(encodeutils.exception_to_unicode(e))
        raise

    return old_index


def delete_index(index_name):
    """Delete the specified index. """

    if index_name is None:
        return
    # Alias will be cleaned up automatically by ES when the index is deleted.
    es_engine = searchlight.elasticsearch.get_api()
    es_engine.indices.delete(index=index_name, ignore=404)


def alias_error_cleanup(indexes):
    """While trying to re-index, we ran into some error. In this case, the
       new index creation/alias updating is incorrect. We will need to clean
       up by rolling back all of the changes. ElasticSearch must stay
       uncluttered. We will delete the indexes explicitly here. ElasticSearch
       will implicitly take care of removing deleted indexes from the aliases.
    """

    es_engine = searchlight.elasticsearch.get_api()

    for index in indexes.values():
        try:
            es_engine.indices.delete(index=index, ignore=404)
        except Exception as e:
            msg = {'index': index}
            LOG.error("Index [%(index)s] clean-up failed." % msg)
            LOG.error(encodeutils.exception_to_unicode(e))


def get_indices(alias):
    """Return a list of indices associated with the specified alias.
    """
    es_engine = searchlight.elasticsearch.get_api()

    return es_engine.indices.get_alias(name=alias)


def normalize_date_fields(document,
                          created_at='created',
                          updated_at='updated'):
    """Attempt to normalize documents to make it easier for consumers,
    particularly around sorting.
    """
    if created_at and 'created_at' not in document:
        document[u'created_at'] = document[created_at]
    if updated_at and 'updated_at' not in document:
        document[u'updated_at'] = document[updated_at]


def get_facets_query(fields, nested_fields, limit_terms):
    term_aggregations = {}
    for facet in fields:
        if isinstance(facet, tuple):
            facet_name, actual_field = facet
        else:
            facet_name, actual_field = facet, facet
        if '.' in facet_name and facet_name.split('.')[0] in nested_fields:
            # Needs a nested aggregate
            term_aggregations[facet_name] = {
                "nested": {"path": facet_name.split('.')[0]},
                "aggs": {
                    # TODO(sjmc7): Handle deeper nesting?

                    facet_name: {
                        'terms': {
                            'field': actual_field,
                            'size': limit_terms
                        },
                        "aggs": {
                            facet_name + '__unique_docs': {
                                "reverse_nested": {}
                            }
                        }
                    }
                }
            }
        else:
            term_aggregations[facet_name] = {
                'terms': {'field': actual_field, 'size': limit_terms}
            }
    return term_aggregations


def transform_facets_results(result_aggregations, resource_type):
    """This effectively reverses the logic from `get_facets_query`,
    and produces output that looks the same regardless of whether
    a faceted field happens to be nested.

    The input should be the value of the `aggregations` keypair in the
    Elasticsearch response. Inputs can be of two forms:
      {"not_nested": {"buckets": [{"key": "something", "doc_count": 10}..]},
       "nested": "buckets": [
          {"key": 4, "doc_count": 2, "nested__unique_docs": {"doc_count": 1}},
          {"key": 6, "doc_count": 3, "nested__unique_docs": {"doc_count": 2}}
      ]}

    Output is normalized (using the __unique_docs count) to:
      {"not_nested": {"buckets": [{"key": "something", "doc_count": 10}..]},
       "nested": {"buckets": [
          {"key": 4, "doc_count": 1},
          {"key": 6, "doc_count": 2}
       ]}
    """
    facet_terms = {}
    for term, aggregation in result_aggregations.items():
        if term in aggregation:
            # Again, deeper nesting question
            term_buckets = aggregation[term]['buckets']
            for bucket in term_buckets:
                reversed_agg = bucket.pop(term + "__unique_docs")
                bucket["doc_count"] = reversed_agg["doc_count"]
            facet_terms[term] = term_buckets
        elif 'buckets' in aggregation:
            facet_terms[term] = aggregation['buckets']
        else:
            # This can happen when there's no mapping defined at all..
            format_msg = {
                'field': term,
                'resource_type': resource_type
            }
            LOG.warning(
                "Unexpected aggregation structure for field "
                "'%(field)s' in %(resource_type)s. Is the mapping "
                "defined correctly?" % format_msg)
            facet_terms[term] = []
    return facet_terms


def find_missing_types(index_type_mapping):
    """Find if doc types are not exist in given indices"""
    missing_index, missing_type = [], []

    if not index_type_mapping:
        return missing_index, missing_type

    es_engine = searchlight.elasticsearch.get_api()

    for index in six.iterkeys(index_type_mapping):
        for doc_type in index_type_mapping[index]:
            try:
                mapping = es_engine.indices.get_mapping(index, doc_type)
                if not mapping:
                    missing_type.append(doc_type)
            except es_exc.NotFoundError:
                missing_index.append(index)

    return set(missing_index), set(missing_type)


def normalize_es_document(es_doc, plugin):
    """
    Remove any elasticsearch specific field and apply plugin's filter_result
    method to given document.
    """
    # Remove user role field
    es_doc.pop(searchlight.elasticsearch.ROLE_USER_FIELD, None)

    # Apply plugin's filter_result
    admin_context = RequestContext()
    plugin.filter_result({'_source': es_doc}, admin_context)
    return es_doc


def check_notification_version(expected, actual, notification_type):
    """
    If actual's major version is different from expected, a
    VersionedNotificationMismatch error is raised.
    If the minor versions are different, a DEBUG level log
    message is output
    """
    maj_ver, min_ver = map(int, actual.split('.'))
    expected_maj, expected_min = map(int, expected.split('.'))
    if maj_ver != expected_maj:
        raise sl_exc.VersionedNotificationMismatch(
            provided_maj=maj_ver, provided_min=min_ver,
            expected_maj=expected_maj, expected_min=expected_min,
            type=notification_type)

    if min_ver != expected_min:
        LOG.debug(
            "Notification minor version mismatch. "
            "Provided: %(provided_maj)s, %(provided_min)s. "
            "Expected: %(expected_maj)s.%(expected_min)s." % {
                "provided_maj": maj_ver, "provided_min": min_ver,
                "expected_maj": expected_maj, "expected_min": expected_min}
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import searchlight.elasticsearch.plugins.ironic as ironic_plugin
from searchlight.elasticsearch.plugins.ironic import ports as ports_plugin
from searchlight.elasticsearch.plugins.ironic import resources as ir_resources
import searchlight.tests.utils as test_utils

PORT_UUID = "eaaca217-e7d8-47b4-bb41-3f99f20eed89"


def _create_port_fixture():
    port = {
        "address": "77:66:23:34:11:b7",
        "created_at": "2016-02-11T15:23:03+00:00",
        "node_uuid": "5b236cab-ad4e-4220-b57c-e827e858745a",
        "extra": {},
        "local_link_connection": {},
        "pxe_enabled": True,
        "updated_at": "2016-03-27T20:41:03+00:00",
        "uuid": PORT_UUID
    }

    return port


class TestPortLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestPortLoaderPlugin, self).setUp()
        self.plugin = ports_plugin.PortIndex()

    def test_default_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Ironic::Port',
                         self.plugin.get_document_type())

    def test_rbac_filter(self):
        rbac = self.plugin._get_rbac_field_filters({})
        self.assertEqual([], rbac)

    def test_admin_only_fields(self):
        admin_only_fields = self.plugin.admin_only_fields
        self.assertEqual([], admin_only_fields)

    def test_document_id(self):
        self.assertEqual('uuid', self.plugin.get_document_id_field())

    def test_facets_with_options(self):
        expected = 'pxe_enabled',
        self.assertEqual(expected, self.plugin.facets_with_options)

    def test_serialize(self):
        serialized = ironic_plugin.serialize_resource(_create_port_fixture(),
                                                      ir_resources.PORT_FIELDS)
        # id cloned from uuid
        self.assertEqual(PORT_UUID, serialized['id'])
        # if name is not set it's uuid
        self.assertEqual(PORT_UUID, serialized['name'])



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from copy import deepcopy
import novaclient.exceptions
from oslo_config import cfg
from oslo_log import log as logging

from elasticsearch import helpers
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.nova import serialize_nova_flavor
from searchlight.elasticsearch.plugins.nova import serialize_nova_server
from searchlight.elasticsearch.plugins.nova import serialize_server_versioned
from searchlight.elasticsearch.plugins import utils
from searchlight import pipeline

LOG = logging.getLogger(__name__)

SERVERGROUP_RETRIES = 20


class InstanceHandler(base.NotificationBase):
    """Handles nova server notifications. These can come as a result of
    a user action (like a name change, state change etc) or as a result of
    periodic auditing notifications nova sends. Because there are so many
    state change notifications, and because currently we're forced to go to
    the nova API for updates, some notifications are processed only as updates
    to the indexed elasticsearch data. Right now those events are part of the
    following operations: boot, poweron/off, suspend/resume, pause/unpause,
                          delete
    """
    partial_state_updates = True

    _state_fields = {'state': 'OS-EXT-STS:vm_state',
                     'new_task_state': 'OS-EXT-STS:task_state'}

    _provisioning_states = {'networking': None,
                            'block_device_mapping': 'networking',
                            'spawning': 'block_device_mapping'}
    _reboot_states = {
        # Hard reboot
        'reboot_pending_hard': 'rebooting_hard',
        'reboot_started_hard': 'reboot_pending_hard',
        # Soft reboot
        'reboot_pending': 'rebooting',
        'reboot_started': 'reboot_pending',
    }
    _shelve_states = {
        'shelving_image_pending_upload': 'shelving',
        'shelving_image_uploading': 'shelving_image_pending_upload',
        # shelving_image_uploading -> shelving_image_pending_upload and
        # shelving_image_pending_upload -> shelving_image_uploading will
        # be ignored.
        'shelving_offloading': 'shelving_image_uploading',
    }
    _unshelve_states = {
        'spawning': 'unshelving'
    }

    # Supported major/minor notification versions. Major changes will likely
    # require code changes.
    notification_versions = {
        'InstanceActionPayload': '1.2',
        'InstanceUpdatePayload': '1.3',
        'InstanceActionVolumeSwapPayload': '1.1',
    }

    @classmethod
    def _get_notification_exchanges(cls):
        return ['nova']

    def get_log_fields(self, event_type, payload):
        return (('id', payload.get('instance_id')),
                ('state', payload.get('state')),
                ('state_description', payload.get('state_description')),
                ('old_task_state', payload.get('old_task_state')),
                ('new_task_state', payload.get('new_task_state')))

    @classmethod
    def get_plugin_opts(cls):
        opts = super(InstanceHandler, cls).get_plugin_opts()
        opts.extend([
            cfg.BoolOpt(
                'use_versioned_notifications',
                help='Expect versioned notifications and ignore unversioned',
                default=True)
        ])
        return opts

    def _use_versioned_notifications(self):
        return self.plugin_options.use_versioned_notifications

    def get_event_handlers(self):
        if not self._use_versioned_notifications():
            return {
                # compute.instance.update seems to be the event set as a
                # result of a state change etc
                'compute.instance.update': self.index_from_update,

                'compute.instance.create.start': self.index_from_api,
                'compute.instance.create.end': self.index_from_api,

                'compute.instance.power_on.end': self.index_from_api,
                'compute.instance.power_off.end': self.index_from_api,
                'compute.instance.resume.end': self.index_from_api,
                'compute.instance.suspend.end': self.index_from_api,
                'compute.instance.pause.end': self.index_from_api,
                'compute.instance.unpause.end': self.index_from_api,

                'compute.instance.shutdown.end': self.index_from_api,
                'compute.instance.reboot.end': self.index_from_api,
                'compute.instance.delete.end': self.delete,

                'compute.instance.shelve.end': self.index_from_api,
                'compute.instance.shelve_offload.end': self.index_from_api,
                'compute.instance.unshelve.end': self.index_from_api,

                'compute.instance.volume.attach': self.index_from_api,
                'compute.instance.volume.detach': self.index_from_api,

                # Removing neutron port events for now; waiting on nova
                # to implement interface notifications as with volumes
                # https://launchpad.net/bugs/1567525
                # bps/nova/+spec/interface-notifications
            }
        # Otherwise listen for versioned notifications!
        # Nova versioned notifications all include the entire payload
        end_events = ['create', 'pause', 'power_off', 'power_on',
                      'reboot', 'rebuild', 'resize', 'restore', 'resume',
                      'shelve', 'shutdown', 'snapshot', 'suspend', 'unpause',
                      'unshelve', 'volume_attach', 'volume_detach']
        notifications = {('instance.%s.end' % ev): self.index_from_versioned
                         for ev in end_events}

        # instance.update has no start or end
        notifications['instance.update'] = self.index_from_versioned

        # This should become soft delete once that is supported
        notifications['instance.delete.end'] = self.delete_from_versioned
        return notifications

    def index_from_update(self, event_type, payload, timestamp):
        """Determine whether or not to process a full update. The updates, and
        how they are processed, are:
        BUILD (state=building unless noted)
         * new_task_state == old_task_state == scheduling -> full index
         * new_task_state == None, old_task_state == scheduling -> ignore
         * new_task_state == old_task_state == None -> ignore
         * new_task_state == networking -> state update
         * new_task_state == block_device_mapping -> state update
         * new_task_state = spawning -> state update
         * state == active, old_task_state == spawning -> ignore
         * state == error -> full index

        There are a set of power state transitions that all follow the same
        kind of structure, with 2 updates and .start and .end events. For these
        events the first update can be a state change and the second ignored.
        They are:
          active -> powering-off -> stopped -> powering-on -> active
          active -> pausing -> paused -> unpausing -> active
          active -> suspending -> suspended -> resuming -> active
        """
        new_state = payload.get('state', None)
        old_state = payload.get('old_state', None)
        new_task_state = payload.get('new_task_state', None)
        old_task_state = payload.get('old_task_state', None)

        # Map provisioning new_task_states to the previous one

        # Default to updating from the API unless we decide otherwise
        update_if_state_matches = None
        ignore_update = False

        if new_state == 'error':
            # There are several ways in which an instance can end up in an
            # error state; it may result in duplicate API requests but it's
            # hard to predict them all
            pass
        elif (new_state == 'active' and old_state == 'active' and
                old_task_state is None and new_task_state is None):
            # This is probably a legitimate server update and should
            # be processed
            pass
        elif new_state == 'building':
            if old_task_state == 'scheduling' and new_task_state is None:
                # This update arrives after scheduling, and immediately prior
                # to create.start (or an update to error state)
                ignore_update = True
            elif new_task_state is None and old_task_state is None:
                # No new state information - this notification isn't useful
                ignore_update = True
            elif new_task_state in self._provisioning_states:
                # Match against the provisioning states above (networking etc)
                update_if_state_matches = dict(
                    state=new_state,
                    new_task_state=self._provisioning_states[new_task_state])
        elif new_state == 'active' and old_task_state == 'spawning':
            # This is received a few microseconds ahead of instance.create.end
            # and indicates the end of the init sequence; ignore this one in
            # favor of create.end
            ignore_update = True
        elif new_task_state is None and old_task_state is not None:
            # These happen right before a corresponding .end event
            ignore_update = True
        elif new_state == 'deleted' and old_task_state == 'deleting':
            # This is always succeeded by a delete.end
            ignore_update = True
        elif new_state == 'active' and new_task_state == 'deleting':
            update_if_state_matches = dict(
                state='active',
                new_task_state=''
            )
        elif new_task_state is not None and new_task_state == old_task_state:
            # Ignore spawning -> spawning for shelved_offloaded instance and
            # shelving_offloading -> shelving_offloading for shelved instance.
            if (new_task_state == 'shelving_offloading' or
                    (new_state == 'shelved_offloaded' and
                     new_task_state == 'spawning')):
                ignore_update = True
            else:
                update_if_state_matches = dict(
                    state=new_state,
                    new_task_state=None)
        elif new_task_state in self._reboot_states:
            update_if_state_matches = dict(
                state=new_state,
                new_task_state=self._reboot_states[new_task_state])
        elif new_task_state in self._shelve_states:
            if (new_state == 'active' or
                    (new_state == 'shelved' and
                     new_task_state == 'shelving_offloading')):
                update_if_state_matches = dict(
                    state='active',
                    new_task_state=self._shelve_states[new_task_state])
        elif new_state == 'shelved_offloaded':
            if new_task_state is not None:
                update_if_state_matches = dict(
                    state=new_state,
                    new_task_state=self._unshelve_states[new_task_state])
            else:
                update_if_state_matches = dict(
                    state='shelved',
                    new_task_state='shelving_offloading')

        if ignore_update:
            LOG.debug("Skipping update or indexing for %s; event contains no "
                      "useful information", payload['instance_id'])
        elif not update_if_state_matches:
            return self.index_from_api(event_type, payload, timestamp)
        elif self.partial_state_updates:
            return self._partial_state_update(
                event_type, payload, update_if_state_matches)
        else:
            LOG.debug("Skipping partial state update for %s; functionality "
                      "is disabled", payload['instance_id'])

    def _partial_state_update(
            self, event_type, payload, update_if_state_matches):
        """Issue a partial document update that will only affect
        state and task_state fields.
        """
        instance_id = payload['instance_id']
        LOG.debug("Attempting partial state update for %s matching %s",
                  instance_id, update_if_state_matches)

        state_field_values = {
            self._state_fields[k]: payload[k]
            for k in self._state_fields if k in payload
        }

        if state_field_values:
            # Don't want to use groovy scripting
            # because there's a high chance it'll be disabled; instead, will
            # get 'n' retrieve ourselves
            preexisting = self.index_helper.get_document(instance_id,
                                                         for_admin=True)

            def should_update(_source):
                for key, value in update_if_state_matches.items():
                    key = self._state_fields[key]
                    if key not in _source:
                        LOG.debug("Skipping state update for %s; precondition "
                                  "'%s' not in existing source",
                                  instance_id, key)
                        return False
                    if _source[key] != value:
                        LOG.debug(
                            "Skipping state update for %s; precondition "
                            "'%s' = '%s' doesn't match '%s' in source",
                            instance_id, key, value, _source[key])
                        return False
                return True

            if preexisting:
                if should_update(preexisting['_source']):
                    LOG.debug("Performing state update for %s", instance_id)
                    # All preconditions matched; save_document will attempt
                    # to save merged document
                    # TODO(sjmc7) - use the existing update_at to generate
                    # a new timestamp? Still seems kind of made up
                    preexisting['_source'].update(state_field_values)
                    self.index_helper.save_document(
                        preexisting['_source'],
                        version=preexisting['_version'] + 1
                    )
                    return pipeline.IndexItem(self.index_helper.plugin,
                                              event_type,
                                              payload,
                                              preexisting['_source'])

    def index_from_api(self, event_type, payload, timestamp):
        """Index from the nova API"""
        instance_id = payload['instance_id']
        LOG.debug("Updating nova server information for %s", instance_id)
        try:
            serialized_payload = serialize_nova_server(instance_id)
            self.index_helper.save_document(
                serialized_payload,
                version=self.get_version(serialized_payload, timestamp))
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      serialized_payload)
        except novaclient.exceptions.NotFound:
            LOG.warning("Instance %s not found; deleting" % instance_id)

            # Where a notification represents an in-progress delete, we will
            # also receive an 'instance.delete' notification shortly
            deleted = (payload.get('state_description') == 'deleting' or
                       payload.get('state') == 'deleted')
            if not deleted:
                return self.delete(event_type, payload, timestamp)

    def delete(self, event_type, payload, timestamp):
        instance_id = payload['instance_id']
        LOG.debug("Deleting nova instance information for %s", instance_id)
        if not instance_id:
            return

        try:
            version = self.get_version(payload, timestamp,
                                       preferred_date_field='deleted_at')
            self.index_helper.delete_document(
                {'_id': instance_id, '_version': version})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       instance_id
                                       )
        except Exception as exc:
            LOG.error(
                'Error deleting instance %(instance_id)s '
                'from index: %(exc)s' %
                {'instance_id': instance_id, 'exc': exc})

    def index_from_versioned(self, event_type, payload, timestamp):
        notification_version = payload['nova_object.version']
        notification_name = payload['nova_object.name']
        expected_version = self.notification_versions.get(notification_name,
                                                          None)
        if expected_version:
            utils.check_notification_version(
                expected_version, notification_version, notification_name)
        else:
            LOG.warning("No expected notification version for %s; "
                        "processing anyway", notification_name)

        versioned_payload = payload['nova_object.data']
        serialized = serialize_server_versioned(versioned_payload)
        self.index_helper.save_document(
            serialized,
            version=self.get_version(serialized, timestamp))
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  serialized)

    def delete_from_versioned(self, event_type, payload, timestamp):
        payload = payload['nova_object.data']
        instance_id = payload['uuid']
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        try:
            version = self.get_version(payload, timestamp,
                                       preferred_date_field='deleted_at')
            self.index_helper.delete_document(
                {'_id': instance_id, '_version': version})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       instance_id
                                       )
        except Exception as exc:
            LOG.error(
                'Error deleting instance %(instance_id)s '
                'from index: %(exc)s' %
                {'instance_id': instance_id, 'exc': exc})


class ServerGroupHandler(base.NotificationBase):
    """Handles nova server group notifications.
    """

    @classmethod
    def _get_notification_exchanges(cls):
        return ['nova']

    def get_event_handlers(self):
        return {
            'servergroup.delete': self.delete,
            'servergroup.create': self.create,
            'servergroup.addmember': self.addmember,
            'compute.instance.delete.end': self.delete_instance
        }

    def _update_server_group_members(self, sg_id, member_id, delete=False):
        # The issue here is that the notification is not complete.
        # We have only a single member that needs to be added to an
        # existing group. A major issue is that we may be updating
        # the ES document while other workers are modifying the rules
        # in the same ES document. This requires an aggressive retry policy,
        # using the "version" field. Since the ES document will have been
        # modified after a conflict, we will need to grab the latest version
        # of the document before continuing. After "retries" number of times,
        # we will admit failure and not try the update anymore.
        LOG.debug("Updating server group member information for %s", sg_id)

        for attempts in range(SERVERGROUP_RETRIES):
            # Read, modify, write of an existing security group.
            doc = self.index_helper.get_document(sg_id)

            if not doc:
                return
            body = doc['_source']
            if not body or 'members' not in body:
                return

            if delete:
                body['members'] = list(filter(
                    lambda r: r != member_id, body['members']))
            else:
                body['members'].append(member_id)

            version = doc['_version']
            try:
                version += 1
                self.index_helper.save_document(body, version=version)
                return body
            except helpers.BulkIndexError as e:
                if e.errors[0]['index']['status'] == 409:
                    # Conflict error, retry with new version of doc.
                    pass
                else:
                    raise

        if attempts == (SERVERGROUP_RETRIES - 1):
            LOG.error('Error updating server group member %(id)s:'
                      ' Too many retries' % {'id': member_id})

    def create(self, event_type, payload, timestamp):
        server_group = deepcopy(payload)
        server_group['id'] = server_group.pop('server_group_id')
        server_group['metadata'] = {}
        server_group['members'] = []
        server_group['updated_at'] = utils.get_now_str()

        LOG.debug("creating nova server group "
                  "information for %s", server_group['id'])
        version = self.get_version(server_group, timestamp)
        self.index_helper.save_document(server_group, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  server_group)

    def addmember(self, event_type, payload, timestamp):
        server_group_id = payload['server_group_id']
        instance_id = payload['instance_uuids'][0]
        server_group = self._update_server_group_members(server_group_id,
                                                         instance_id)
        if server_group:
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      server_group)

    def delete_instance(self, event_type, payload, timestamp):

        # When an instance is deleted from Nova, its' record in
        # InstanceGroup DB was cleaned directly from DB layer in
        # Nova, we should perform sync to keep Searchlight
        # up-to-date with Nova DB.
        instance_id = payload['instance_id']

        query = {'filter': {'term': {'members': instance_id}}}
        search_results = self.index_helper.simple_search(
            query=query, type="OS::Nova::ServerGroup")

        server_group = None
        try:
            result = search_results['hits'][0]
            server_group_id = result['_id']
            server_group = self._update_server_group_members(
                server_group_id,
                instance_id,
                delete=True)
        except IndexError:
            LOG.debug("No nova server group information for instance %s",
                      instance_id)
        if server_group:
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      server_group)

    def delete(self, event_type, payload, timestamp):
        server_group_id = payload['server_group_id']
        LOG.debug("Deleting nova server group information for %s",
                  server_group_id)
        if not server_group_id:
            return

        try:
            self.index_helper.delete_document({'_id': server_group_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       server_group_id
                                       )
        except Exception as exc:
            LOG.error(
                'Error deleting server group %(server_group_id)s '
                'from index: %(exc)s' %
                {'server_group_id': server_group_id, 'exc': exc})


class FlavorHandler(base.NotificationBase):
    """Handles nova flavor versioned notifications. The payload samples are:
       https://docs.openstack.org/nova/latest/reference/notifications.html#existing-versioned-notifications
"""

    @classmethod
    def _get_notification_exchanges(cls):
        return ['nova']

    def get_event_handlers(self):
        return {
            'flavor.create': self.create_or_update,
            'flavor.update': self.create_or_update,
            'flavor.delete': self.delete
        }

    def get_log_fields(self, event_type, payload):
        return ('flavorid', payload['nova_object.data']['flavorid']),

    def create_or_update(self, event_type, payload, timestamp):
        flavor = serialize_nova_flavor(payload['nova_object.data'])
        version = self.get_version(flavor, timestamp)

        LOG.debug("Updating nova flavor information for %s", flavor['id'])
        self.index_helper.save_document(flavor, version=version)
        return pipeline.IndexItem(self.index_helper.plugin,
                                  event_type,
                                  payload,
                                  flavor)

    def delete(self, event_type, payload, timestamp):
        flavor = payload['nova_object.data']
        flavor_id = flavor['flavorid']
        LOG.debug("Deleting nova flavor information for %s", flavor_id)

        try:
            self.index_helper.delete_document({'_id': flavor_id})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       flavor_id)
        except Exception as exc:
            LOG.error('Error deleting flavor %(flavor_id)s '
                      'from index: %(exc)s' %
                      {'flavor_id': flavor_id, 'exc': exc})



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import six

from searchlight.elasticsearch.plugins import openstack_clients

LOG = logging.getLogger(__name__)


BLACKLISTED_FIELDS = set((u'links', u'manager', '_loaded', '_info'))


def serialize_cinder_volume(volume):
    """volume can be an id or a 'volume' object from cinderclient"""
    if isinstance(volume, six.string_types):
        cinder_client = openstack_clients.get_cinderclient()
        volume = cinder_client.volumes.get(volume)

    LOG.debug("Serializing volume %s for project %s",
              volume.id, volume.user_id)

    serialized = {k: v for k, v in volume.to_dict().items()
                  if k not in BLACKLISTED_FIELDS}

    project_id = serialized.get('os-vol-tenant-attr:tenant_id')
    if 'tenant_id' not in serialized:
        serialized['tenant_id'] = project_id
    if 'project_id' not in serialized:
        serialized['project_id'] = project_id

    return serialized


def serialize_cinder_snapshot(snapshot):
    """snapshot can be an id or a 'Snapshot' object from cinderclient"""
    if isinstance(snapshot, six.string_types):
        cinder_client = openstack_clients.get_cinderclient()
        snapshot = cinder_client.volume_snapshots.get(snapshot)

    project_id = getattr(snapshot,
                         'os-extended-snapshot-attributes:project_id')
    LOG.debug("Serializing snapshot %s for project %s",
              snapshot.id, project_id)

    serialized = {k: v for k, v in snapshot.to_dict().items()
                  if k not in BLACKLISTED_FIELDS}

    if 'tenant_id' not in serialized:
        serialized['tenant_id'] = project_id
    if 'project_id' not in serialized:
        serialized['project_id'] = project_id

    return serialized



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from searchlight.api.middleware import context
import searchlight.context
import searchlight.tests.utils as test_utils


class TestContextMiddleware(test_utils.BaseTestCase):
    def test_response(self):
        middleware = context.ContextMiddleware(None)
        req = webob.Request.blank('/')
        req.context = searchlight.context.RequestContext()
        request_id = req.context.request_id

        resp = webob.Response()
        resp.request = req
        middleware.process_response(resp)
        self.assertEqual(request_id, resp.headers['x-openstack-request-id'])
        resp_req_id = resp.headers['x-openstack-request-id']
        # Validate that request-id do not starts with 'req-req-'
        self.assertFalse(resp_req_id.startswith('req-req-'))
        self.assertTrue(resp_req_id.startswith('req-'))

    def test_is_admin_project(self):
        middleware = context.ContextMiddleware(None)
        req = webob.Request.blank('/')
        req_context = middleware._get_authenticated_context(req)
        self.assertTrue(req_context.is_admin_project)

        req = webob.Request.blank('/', headers={'X-Is-Admin-Project': 'True'})
        req_context = middleware._get_authenticated_context(req)
        self.assertTrue(req_context.is_admin_project)

        req = webob.Request.blank('/', headers={'X-Is-Admin-Project': 'False'})
        req_context = middleware._get_authenticated_context(req)
        self.assertFalse(req_context.is_admin_project)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_service import service as os_service

from searchlight import listener
from searchlight import service


CONF = cfg.CONF
CONF.import_group("listener", "searchlight.listener")


def main():
    service.prepare_service()
    launcher = os_service.ProcessLauncher(CONF, restart_method='mutate')
    launcher.launch_service(
        listener.ListenerService(),
        workers=CONF.listener.workers)
    launcher.wait()


if __name__ == "__main__":
    main()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Searchlight exception subclasses"""

import six

from searchlight.i18n import _

_FATAL_EXCEPTION_FORMAT_ERRORS = False


class SearchlightException(Exception):
    """
    Base Searchlight Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = _("An unknown exception occurred")

    def __init__(self, message=None, *args, **kwargs):
        if not message:
            message = self.message
        try:
            if kwargs:
                message = message % kwargs
        except Exception:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise
            else:
                # at least get the core message out if something happened
                pass
        self.msg = message
        super(SearchlightException, self).__init__(message)

    def __unicode__(self):
        # NOTE(flwang): By default, self.msg is an instance of Message, which
        # can't be converted by str(). Based on the definition of
        # __unicode__, it should return unicode always.
        return six.text_type(self.msg)


class NotFound(SearchlightException):
    message = _("An object with the specified identifier was not found.")


class Duplicate(SearchlightException):
    message = _("An object with the same identifier already exists.")


class Forbidden(SearchlightException):
    message = _("You are not authorized to complete this action.")


class Invalid(SearchlightException):
    message = _("Data supplied was not valid.")


class InvalidPropertyProtectionConfiguration(Invalid):
    message = _("Invalid configuration in property protection file.")


class ReservedProperty(Forbidden):
    message = _("Attribute '%(property)s' is reserved.")


class InvalidContentType(SearchlightException):
    message = _("Invalid content type %(content_type)s")


class WorkerCreationFailure(SearchlightException):
    message = _("Server worker creation failed: %(reason)s.")


class SchemaLoadError(SearchlightException):
    message = _("Unable to load schema: %(reason)s")


class InvalidObject(SearchlightException):
    message = _("Provided object does not match schema "
                "'%(schema)s': %(reason)s")


class SIGHUPInterrupt(SearchlightException):
    message = _("System SIGHUP signal received.")


class JsonPatchException(SearchlightException):
    message = _("Invalid jsonpatch request")


class InvalidJsonPatchBody(JsonPatchException):
    message = _("The provided body %(body)s is invalid "
                "under given schema: %(schema)s")


class InvalidJsonPatchPath(JsonPatchException):
    message = _("The provided path '%(path)s' is invalid: %(explanation)s")

    def __init__(self, message=None, *args, **kwargs):
        self.explanation = kwargs.get("explanation")
        super(InvalidJsonPatchPath, self).__init__(message, *args, **kwargs)


class IndexingException(SearchlightException):
    message = _("An error occurred during index creation or initial loading")


class InvalidAPIVersionProvided(SearchlightException):
    message = _("The provided API version is not supported, "
                "the current available version range for %(service)s "
                "is: from %(min_version)s to %(max_version)s.")


class VersionedNotificationMismatch(SearchlightException):
    message = _("Provided notification version "
                "%(provided_maj)s.%(provided_min)s did not match expected "
                "%(expected_maj)s.%(expected_min)s for %(type)s")



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

from oslo_config import cfg
import oslo_utils

from searchlight.cmd import manage
from searchlight.elasticsearch.plugins import utils as es_utils
from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight.tests import fake_plugins
from searchlight.tests import functional
from searchlight.tests import utils as test_utils


now = oslo_utils.timeutils.utcnow()
now_str = now.strftime(es_utils.FORMAT)


class TestSearchLoad(functional.FunctionalTest):
    def setUp(self):
        super(TestSearchLoad, self).setUp()
        role_plugin = fake_plugins.RoleSeparatedPlugin(
            es_engine=self.elastic_connection)
        self.role_plugin = role_plugin

        non_role_plugin = fake_plugins.NonRoleSeparatedPlugin(
            es_engine=self.elastic_connection)

        self.non_role_plugin = non_role_plugin

        self.initialized_plugins['non-role-separated'] = non_role_plugin
        self.initialized_plugins['role-separated'] = role_plugin

    def test_role_separation(self):
        index_name = self.role_plugin.alias_name_listener
        doc_type = self.role_plugin.document_type

        self.role_plugin.index_initial_data()
        self._flush_elasticsearch(index_name)

        # Should be 4 documents from the original two source ones
        self.assertEqual(2, self.role_plugin.number_documents)
        es_results = self._get_all_elasticsearch_docs()

        self.assertEqual(
            set(['role-fake1_ADMIN', 'role-fake2_ADMIN',
                 'role-fake1_USER', 'role-fake2_USER']),
            set(hit['_id'] for hit in es_results['hits']['hits']))

        admin1_doc = self._get_elasticsearch_doc(
            index_name, doc_type, 'role-fake1_ADMIN'
        )
        self.assertIn('admin_specific', admin1_doc['_source'])
        self.assertIn('admin_wildcard_this', admin1_doc['_source'])
        self.assertIn('public_field', admin1_doc['_source'])
        self.assertEqual('admin',
                         admin1_doc['_source'][ROLE_USER_FIELD])

        user1_doc = self._get_elasticsearch_doc(
            index_name, doc_type, 'role-fake1_USER'
        )
        self.assertNotIn('admin_specific', user1_doc['_source'])
        self.assertNotIn('admin_wildcard_this', user1_doc['_source'])
        self.assertIn('public_field', user1_doc['_source'])
        self.assertEqual('admin',
                         admin1_doc['_source'][ROLE_USER_FIELD])

    def test_non_separated(self):
        index_name = self.non_role_plugin.alias_name_listener

        self.non_role_plugin.index_initial_data()
        self._flush_elasticsearch(index_name)

        self.assertEqual(2, self.non_role_plugin.number_documents)
        es_results = self._get_all_elasticsearch_docs()
        es_hits = self._get_hit_source(es_results)

        self.assertEqual(2, len(es_hits))
        self.assertEqual(
            set(['non-role-fake1', 'non-role-fake2']),
            set(hit['_id'] for hit in es_results['hits']['hits']))

        self.assertEqual(
            ['admin', 'user'], sorted(es_hits[0][ROLE_USER_FIELD]))

    def test_index_settings(self):
        """Test the default gc_delete interval plus some other
        dynamic index settings
        """
        with mock.patch.object(cfg.CONF, 'elasticsearch') as mock_settings:
            mock_settings.index_gc_deletes = '100s'
            mock_settings.index_settings = {
                'refresh_interval': '2s',
                'index.number_of_replicas': 1
            }

            index_name = es_utils.create_new_index('test-index-settings')
            try:
                settings = self.elastic_connection.indices.get_settings(
                    index_name)
                index_settings = settings[index_name]['settings']['index']

                self.assertEqual("100s", index_settings['gc_deletes'])
                self.assertEqual("2s", index_settings['refresh_interval'])
                self.assertEqual("1", index_settings['number_of_replicas'])

            finally:
                es_utils.delete_index(index_name)

    @mock.patch('oslo_utils.timeutils.utcnow')
    @mock.patch('searchlight.common.utils.get_search_plugins')
    def test_manage(self, mock_get_plugins, mock_utcnow):
        """Test that manage index sync works from end to end. Uses fake plugins
        because it avoids having to fake service data and is less dependent
        on functional tests for each service plugin.
        """
        mock_utcnow.return_value = datetime.datetime(year=2016, month=1, day=1)
        expected_index_name = 'searchlight-2016_01_01_00_00_00'

        simple_plugin = fake_plugins.FakeSimplePlugin(self.elastic_connection)
        child_plugin = fake_plugins.FakeChildPlugin(self.elastic_connection)
        non_role_plugin = fake_plugins.NonRoleSeparatedPlugin(
            self.elastic_connection)
        child_plugin.register_parent(simple_plugin)

        mock_get_plugins.return_value = {
            plugin.get_document_type(): test_utils.StevedoreMock(plugin)
            for plugin in (simple_plugin, child_plugin, non_role_plugin)
        }
        index_command = manage.IndexCommands()
        # The fake plugins all have hardcoded data for get_objects that will
        # be indexed. Use force=True to avoid the 'are you sure?' prompt
        try:
            index_command.sync(force=True)

            es_results = self._get_all_elasticsearch_docs()
            es_hits = self._get_hit_source(es_results)
        finally:
            es_utils.delete_index(expected_index_name)

        self.assertEqual(expected_index_name,
                         es_results['hits']['hits'][0]['_index'])

        expected = ['simple1', 'child1', 'non-role-fake1', 'non-role-fake2']
        self.assertEqual(len(expected), len(es_hits))
        self.assertEqual(
            set(expected),
            set(hit['_id'] for hit in es_results['hits']['hits']))

    @mock.patch('oslo_utils.timeutils.utcnow')
    @mock.patch('searchlight.common.utils.get_search_plugins')
    def test_manage_type(self, mock_get_plugins, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(year=2016, month=2, day=2)
        expected_index_name = 'searchlight-2016_02_02_00_00_00'

        simple_plugin = fake_plugins.FakeSimplePlugin(self.elastic_connection)
        non_role_plugin = fake_plugins.NonRoleSeparatedPlugin(
            self.elastic_connection)

        mock_get_plugins.return_value = {
            plugin.get_document_type(): test_utils.StevedoreMock(plugin)
            for plugin in (simple_plugin, non_role_plugin)
        }
        index_command = manage.IndexCommands()

        # Expect this to match simple plugin and child plugin, but not
        # non_role_plugin. Patch the index->index function call since it won't
        # find any data, and we want to check it's called correctly
        with mock.patch.object(index_command,
                               '_es_reindex_worker') as patch_es_reindex:
            try:
                index_command.sync(_type='fake-simple', force=True)

                es_results = self._get_all_elasticsearch_docs()
                es_hits = self._get_hit_source(es_results)
            finally:
                es_utils.delete_index(expected_index_name)

            patch_es_reindex.assert_called_with(
                {non_role_plugin.get_document_type(): non_role_plugin},
                [('searchlight', 'searchlight-search',
                 'searchlight-listener')],
                {'searchlight': expected_index_name}
            )

        expected = ['simple1']
        self.assertEqual(len(expected), len(es_hits))
        self.assertEqual(
            set(expected),
            set(hit['_id'] for hit in es_results['hits']['hits']))

    @mock.patch('oslo_utils.timeutils.utcnow')
    @mock.patch('searchlight.common.utils.get_search_plugins')
    def test_manage_type_glob(self, mock_get_plugins, mock_utcnow):
        mock_utcnow.return_value = datetime.datetime(year=2016, month=3, day=3)
        expected_index_name = 'searchlight-2016_03_03_00_00_00'

        simple_plugin = fake_plugins.FakeSimplePlugin(self.elastic_connection)
        child_plugin = fake_plugins.FakeChildPlugin(self.elastic_connection)
        non_role_plugin = fake_plugins.NonRoleSeparatedPlugin(
            self.elastic_connection)
        child_plugin.register_parent(simple_plugin)

        mock_get_plugins.return_value = {
            plugin.get_document_type(): test_utils.StevedoreMock(plugin)
            for plugin in (simple_plugin, child_plugin, non_role_plugin)
        }
        index_command = manage.IndexCommands()

        # Expect this to match simple plugin and child plugin, but not
        # non_role_plugin. Patch the index->index function call since it won't
        # find any data, and we want to check it's called correctly
        with mock.patch.object(index_command,
                               '_es_reindex_worker') as patch_es_reindex:
            try:
                # Use two wildcard matches
                index_command.sync(_type='fake-sim*,fake-chi*', force=True)

                es_results = self._get_all_elasticsearch_docs()
                es_hits = self._get_hit_source(es_results)
            finally:
                es_utils.delete_index(expected_index_name)

            patch_es_reindex.assert_called_with(
                {non_role_plugin.get_document_type(): non_role_plugin},
                [('searchlight', 'searchlight-search',
                 'searchlight-listener')],
                {'searchlight': expected_index_name}
            )

        expected = ['simple1', 'child1']
        self.assertEqual(len(expected), len(es_hits))
        self.assertEqual(
            set(expected),
            set(hit['_id'] for hit in es_results['hits']['hits']))

    @mock.patch.object(cfg.CONF, 'service_credentials')
    @mock.patch.object(cfg.CONF, 'resource_plugin')
    def test_without_region(self, resource_plugin_conf, service_cred_conf):
        # Test that region isn't indexed unless explicitly enabled
        resource_plugin_conf.include_region_name = False
        service_cred_conf.os_region_name = 'test-region'
        index_name = self.role_plugin.alias_name_listener
        simple_plugin = fake_plugins.FakeSimplePlugin(self.elastic_connection)

        simple_plugin.index_initial_data()
        self._flush_elasticsearch(index_name)

        es_results = self._get_all_elasticsearch_docs()
        self.assertNotIn('region_name',
                         es_results['hits']['hits'][0]['_source'])

    @mock.patch.object(cfg.CONF, 'service_credentials')
    @mock.patch.object(cfg.CONF, 'resource_plugin')
    def test_with_region(self, resource_plugin_conf, service_cred_conf):
        # Now test with a region
        resource_plugin_conf.include_region_name = True
        service_cred_conf.os_region_name = 'test-region'

        index_name = self.role_plugin.alias_name_listener

        simple_plugin = fake_plugins.FakeSimplePlugin(self.elastic_connection)
        non_role = fake_plugins.NonRoleSeparatedPlugin(self.elastic_connection)

        # Override region name for non-role
        non_role.options.override_region_name = ['region1', 'region2']

        non_role.index_initial_data()
        simple_plugin.index_initial_data()
        self._flush_elasticsearch(index_name)

        es_results = self._get_all_elasticsearch_docs()

        simple_res = list(filter(
            lambda r: r['_type'] == simple_plugin.get_document_type(),
            es_results['hits']['hits']))
        non_role_res = list(filter(
            lambda r: r['_type'] == non_role.get_document_type(),
            es_results['hits']['hits']))

        self.assertEqual(
            ['test-region'],
            simple_res[0]['_source']['region_name'])

        self.assertEqual(['region1', 'region2'],
                         non_role_res[0]['_source']['region_name'])



# Copyright 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from searchlight.api.v1 import search
from searchlight.common import wsgi


class API(wsgi.Router):

    """WSGI router for Searchlight v1 API requests."""

    def __init__(self, mapper):

        reject_method_resource = wsgi.Resource(wsgi.RejectMethodController())

        search_catalog_resource = search.create_resource()
        mapper.connect('/search',
                       controller=search_catalog_resource,
                       action='search',
                       conditions={'method': ['GET']})
        mapper.connect('/search',
                       controller=search_catalog_resource,
                       action='search',
                       conditions={'method': ['POST']})
        mapper.connect('/search',
                       controller=reject_method_resource,
                       action='reject',
                       allowed_methods='GET, POST',
                       conditions={'method': ['PUT', 'DELETE',
                                              'PATCH', 'HEAD']})

        mapper.connect('/search/plugins',
                       controller=search_catalog_resource,
                       action='plugins_info',
                       conditions={'method': ['GET']})
        mapper.connect('/search/plugins',
                       controller=reject_method_resource,
                       action='reject',
                       allowed_methods='GET',
                       conditions={'method': ['POST', 'PUT', 'DELETE',
                                              'PATCH', 'HEAD']})

        mapper.connect('/search/facets',
                       controller=search_catalog_resource,
                       action='facets',
                       conditions={'method': ['GET']}),
        mapper.connect('/search/facets',
                       controller=reject_method_resource,
                       action='reject',
                       allowed_methods='GET',
                       conditions={'method': ['POST', 'PUT', 'DELETE',
                                              'PATCH', 'HEAD']})

        super(API, self).__init__(mapper)



# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
from unittest import mock

from searchlight.elasticsearch.plugins.base import NotificationBase
from searchlight.elasticsearch.plugins.glance import metadefs as md_plugin
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils

# Metadefinitions
USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'

TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'

NAMESPACE1 = 'namespace1'
NAMESPACE2 = 'namespace2'

PROPERTY1 = 'Property1'
PROPERTY2 = 'Property2'
PROPERTY3 = 'Property3'

OBJECT1 = 'Object1'
OBJECT2 = 'Object2'
OBJECT3 = 'Object3'

RESOURCE_TYPE1 = 'ResourceType1'
RESOURCE_TYPE2 = 'ResourceType2'
RESOURCE_TYPE3 = 'ResourceType3'

TAG1 = 'Tag1'
TAG2 = 'Tag2'
TAG3 = 'Tag3'

now = datetime.datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')


# TODO(sjmc7): Move this stuff to a fixtures module. These should match
# responses from the glance v2 API
def _namespace_fixture(**kwargs):
    obj = {
        'created_at': now,
        'updated_at': now,
        'namespace': None,
        'display_name': None,
        'description': None,
        'visibility': 'public',
        'protected': True,
        'owner': None,
        'schema': '/v2/schemas/metadefs/namespace',
        'resource_type_associations': [],
        'tags': [],
        'objects': [],
        'properties': {}
    }
    obj.update(kwargs)
    return obj


def _property_fixture(title, **kwargs):
    obj = {
        'description': None,
        'title': title,
        'type': 'string'
    }
    obj.update(kwargs)
    return obj


def _object_fixture(name, **kwargs):
    obj = {
        'name': name,
        'description': None,
        'properties': {}
    }
    obj.update(kwargs)
    return obj


def _db_resource_type_fixture(name, **kwargs):
    obj = {
        'name': name,
        'protected': False,
    }
    obj.update(kwargs)
    return obj


def _namespace_resource_type_fixture(name, prefix, **kwargs):
    obj = {
        'properties_target': None,
        'prefix': prefix,
        'name': name,
    }
    obj.update(kwargs)
    return obj


def _tag_fixture(name, **kwargs):
    obj = {
        'name': name,
    }
    obj.update(**kwargs)
    return obj


class TestMetadefLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestMetadefLoaderPlugin, self).setUp()

        self._create_namespaces()
        self._create_namespace_resource_types()
        self._create_properties()
        self._create_tags()
        self._create_objects()

        self.plugin = md_plugin.MetadefIndex()

        mock_session = mock.Mock()
        mock_session.get_endpoint.return_value = 'http://localhost/glance/v2'
        patched_ses = mock.patch(
            'searchlight.elasticsearch.plugins.openstack_clients._get_session',
            return_value=mock_session)
        patched_ses.start()
        self.addCleanup(patched_ses.stop)

    def _create_namespaces(self):
        self.namespaces = [
            _namespace_fixture(namespace=NAMESPACE1,
                               display_name='1',
                               description='desc1',
                               visibility='private',
                               protected=True,
                               owner=TENANT1,
                               resource_type_associations=[
                                   {"name": RESOURCE_TYPE1}
                               ]),
            _namespace_fixture(namespace=NAMESPACE2,
                               display_name='2',
                               description='desc2',
                               visibility='public',
                               protected=False,
                               owner=TENANT1,
                               resource_type_associations=[
                                   {"name": RESOURCE_TYPE2},
                                   {"name": RESOURCE_TYPE3}
                               ]),
        ]

    def _create_properties(self):
        properties = [
            _property_fixture('title1'),
            _property_fixture('title2'),
            _property_fixture('title3')
        ]

        self.namespaces[0]['properties'] = {
            PROPERTY1: properties[0]
        }
        self.namespaces[1]['properties'] = {
            PROPERTY2: properties[1],
            PROPERTY3: properties[2]
        }

    def _create_objects(self):
        objects = [
            _object_fixture(name=OBJECT1,
                            description='desc1',
                            properties={
                                'property1': _property_fixture(
                                    'something title',
                                    type='string',
                                    enum=['value1', 'value2'],
                                    default='value1')
                            }),
            _object_fixture(name=OBJECT2,
                            description='desc2'),
            _object_fixture(name=OBJECT3,
                            description='desc3'),
        ]

        self.namespaces[0]['objects'] = objects[:1]
        self.namespaces[1]['objects'] = objects[1:]

    def _create_namespace_resource_types(self):
        namespace_resource_types = [
            _namespace_resource_type_fixture(
                prefix='p1',
                name=RESOURCE_TYPE1),
            _namespace_resource_type_fixture(
                prefix='p2',
                name=RESOURCE_TYPE2),
            _namespace_resource_type_fixture(
                prefix='p2',
                name=RESOURCE_TYPE3),
        ]
        self.namespaces[0]['resource_type_associations'] = \
            namespace_resource_types[:1]
        self.namespaces[1]['resource_type_associations'] = \
            namespace_resource_types[1:]

    def _create_tags(self):
        tags = [
            _tag_fixture(name=TAG1),
            _tag_fixture(name=TAG2),
            _tag_fixture(name=TAG3),
        ]
        self.namespaces[0]['tags'] = tags[:1]
        self.namespaces[1]['tags'] = tags[1:]

    def _get_namespace(self, namespace):
        """The 'get' namespace API call returns everything (objects, properties,
        resource type allocations) whereas 'list' does not.
        """
        if isinstance(namespace, int):
            return self.namespaces[namespace]
        else:
            for n in self.namespaces:
                if n['namespace'] == namespace:
                    return n

    def _list_namespaces(self):
        """Return a stripped down copy of namespaces, minus tags, properties,
        objects, similar to glanceclient.
        """
        for namespace in self.namespaces:
            ns_copy = copy.deepcopy(namespace)
            del ns_copy['tags']
            del ns_copy['properties']
            del ns_copy['objects']
            yield ns_copy

    def test_resource_group_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Glance::Metadef',
                         self.plugin.get_document_type())

    def test_complex_serialize(self):
        expected = {
            'created_at': now,
            'updated_at': now,
            'namespace': 'namespace1',
            'name': '1',
            'display_name': '1',
            'id': 'namespace1',
            'description': 'desc1',
            'visibility': 'private',
            'protected': True,
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'objects': [{
                'description': 'desc1',
                'name': 'Object1',
                'properties': [{
                    'default': 'value1',
                    'description': None,
                    'enum': ['value1', 'value2'],
                    'name': 'property1',
                    'type': 'string',
                    'title': 'something title'
                }]
            }],
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'resource_types': [{
                # TODO(sjmc7): Removing these because i'm not sure we
                # have access to them
                # 'prefix': 'p1',
                'name': 'ResourceType1',
                # 'properties_target': None
            }],
            'properties': [{
                'name': 'Property1',
                'title': 'title1',
                'type': 'string',
                'description': None
            }],
            'tags': ['Tag1'],
        }

        ns = list(self._list_namespaces())[0]
        with mock.patch('glanceclient.v2.metadefs.NamespaceController.get',
                        return_value=self._get_namespace(ns['namespace'])):
            serialized = self.plugin.serialize(ns)
        self.assertEqual(expected, serialized)

    def test_serialize_no_tags(self):
        ns = copy.deepcopy(list(self._list_namespaces())[0])
        return_value = self._get_namespace(0)
        del return_value['tags']

        expected = {
            'created_at': now,
            'updated_at': now,
            'namespace': 'namespace1',
            'name': '1',
            'id': 'namespace1',
            'display_name': '1',
            'description': 'desc1',
            'visibility': 'private',
            'protected': True,
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'objects': [{
                'description': 'desc1',
                'name': 'Object1',
                'properties': [{
                    'default': 'value1',
                    'description': None,
                    'enum': ['value1', 'value2'],
                    'name': 'property1',
                    'type': 'string',
                    'title': 'something title'
                }]
            }],
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'resource_types': [{
                # TODO(sjmc7): Removing these because i'm not sure we
                # have access to them
                # 'prefix': 'p1',
                'name': 'ResourceType1',
                # 'properties_target': None
            }],
            'properties': [{
                'name': 'Property1',
                'title': 'title1',
                'type': 'string',
                'description': None
            }],
            'tags': [],
        }
        with mock.patch('glanceclient.v2.metadefs.NamespaceController.get',
                        return_value=return_value):
            serialized = self.plugin.serialize(ns)
        self.assertEqual(expected, serialized)

    def test_index_initial_data(self):
        with mock.patch.object(self.plugin, 'get_objects',
                               return_value=self.namespaces) as mock_get:
            with mock.patch.object(self.plugin.index_helper,
                                   'save_documents') as mock_save:
                self.plugin.index_initial_data()
                versions = [NotificationBase.get_version(obj)
                            for obj in self.namespaces]
                mock_get.assert_called_once_with()
                mock_save.assert_called_once_with([
                    {
                        'created_at': now,
                        'updated_at': now,
                        'display_name': '1',
                        'description': 'desc1',
                        'id': 'namespace1',
                        'name': '1',
                        'objects': [
                            {
                                'name': 'Object1',
                                'description': 'desc1',
                                'properties': [{
                                    'default': 'value1',
                                    'name': 'property1',
                                    'enum': ['value1', 'value2'],
                                    'type': 'string',
                                    'title': 'something title',
                                    'description': None
                                }],
                            }
                        ],
                        'namespace': 'namespace1',
                        'visibility': 'private',
                        'protected': True,
                        'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
                        'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
                        'properties': [{
                            'name': 'Property1',
                            'type': 'string',
                            'title': 'title1',
                            'description': None
                        }],
                        'resource_types': [{
                            'name': 'ResourceType1',
                        }],
                        'tags': ['Tag1'],
                    },
                    {
                        'created_at': now,
                        'updated_at': now,
                        'display_name': '2',
                        'description': 'desc2',
                        'id': 'namespace2',
                        'name': '2',
                        'objects': [
                            {
                                'properties': [],
                                'name': 'Object2',
                                'description': 'desc2'
                            },
                            {
                                'properties': [],
                                'name': 'Object3',
                                'description': 'desc3'
                            }
                        ],
                        'namespace': 'namespace2',
                        'visibility': 'public',
                        'protected': False,
                        'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
                        'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
                        'properties': [
                            {
                                'name': 'Property2',
                                'type': 'string',
                                'title': 'title2',
                                'description': None
                            },
                            {
                                'name': 'Property3',
                                'type': 'string',
                                'title': 'title3',
                                'description': None
                            }
                        ],
                        'resource_types': [
                            {
                                'name': 'ResourceType2'
                            },
                            {
                                'name': 'ResourceType3'
                            }
                        ],
                        'tags': ['Tag2', 'Tag3']
                    }
                ], index=None, versions=versions)

    def test_metadef_rbac(self):
        """Test metadefs RBAC query terms"""
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search'
        )
        rbac_filter = self.plugin.get_query_filters(fake_request.context)
        expected_fragment = {
            "indices": {
                "index": "searchlight-search",
                "no_match_query": "none",
                "query": {
                    "bool": {
                        "filter": {
                            "bool": {
                                "must": {
                                    "type": {"value": "OS::Glance::Metadef"}
                                },
                                "should": [
                                    {"term": {"owner": TENANT1}},
                                    {"term": {"visibility": "public"}},
                                ],
                                "minimum_should_match": 1
                            }
                        }
                    }
                }
            }
        }
        self.assertEqual(expected_fragment, rbac_filter)

    def test_filter_result(self):
        """We modify 'tags' to make it fit with mappings from other plugins.
        For now, we modify results to more closely fit the API.
        """
        ns = list(self._list_namespaces())[0]
        with mock.patch('glanceclient.v2.metadefs.NamespaceController.get',
                        return_value=self._get_namespace(ns['namespace'])):
            es_result = self.plugin.serialize(ns)

        self.plugin.filter_result({'_source': es_result}, None)
        self.assertEqual([{"name": "Tag1"}], es_result["tags"])
